"""
    PlateNumberEgh.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.Peaks.ElutionModels import egh
from .ColumnConstants import INJECTION_TIME

Ti = INJECTION_TIME
Np = get_setting("num_plates_pc")

def _pre_optimize_params(ecurve, xr_params, debug=False):
    x = ecurve.x
    y = ecurve.y

    n = xr_params.shape[0]
    temp_xr_params = np.zeros((n,3))

    for k, (h, m) in enumerate(xr_params[:,[0,1]]):
        temp_xr_params[k,0] = h
        temp_xr_params[k,1] = m
        temp_xr_params[k,2] = 0

    def objective(p):
        T_, N_ = p
        sqN = np.sqrt(N_)
        my = np.zeros(len(x))
        for k, (h, m, t) in enumerate(temp_xr_params):
            s = (m - T_)/sqN
            my += egh(x, h, m, s, t)
        return np.sum((my - y)**2)

    ret = minimize(objective, (Ti, Np))

    if debug:
        import molass_legacy.KekLib.DebugPlot as plt

        with plt.Dp():
            fig, (ax0, ax1, ax2) = plt.subplots(ncols=3, figsize=(18,5))
            fig.suptitle("Plate Number Adaption Proof", fontsize=20)
            ax0.set_title("Free EGH Decomposition", fontsize=16)
            ax1.set_title("Plate Number Simple Conversion", fontsize=16)
            ax2.set_title("Plate Number Optimized Adaption", fontsize=16)

            ax0.plot(x, y)
            for k, (h, m, s, t) in enumerate(xr_params):
                ax0.plot(x, egh(x, h, m, s, t), ":")

            def plot_params(ax, Ti, Np):
                ax.plot(x, y)
                for k, (h, m, t) in enumerate(temp_xr_params):
                    s = (m - Ti)/np.sqrt(Np)
                    print([k], h, m, s, t)
                    ax.plot(x, egh(x, h, m, s, t), ":")
            plot_params(ax1, Ti, Np)
            plot_params(ax2, *ret.x)

            fig.tight_layout()
            plt.show()

    return temp_xr_params, ret.x

def pn_estimate_initial_params(ecurve, init_xr_params, seccol_params):
    temp_xr_params, (T_, N_) = _pre_optimize_params(ecurve, init_xr_params)
    pn_seccol_params = np.concatenate([seccol_params, [T_, N_]])
    return temp_xr_params, pn_seccol_params

def convert_to_xr_params(c_xr_params, Ti, Np):
    nc = c_xr_params.shape[0]
    xr_params = np.zeros((nc, 4))
    xr_params[:,[0,1,3]] = c_xr_params
    sqrtNp = np.sqrt(Np)
    for k in range(nc):
        xr_params[k,2] = (xr_params[k,1] - Ti)/sqrtNp
    return xr_params

if __name__ == '__main__':
    pass



# coding: utf-8
"""
    Rgg.UvAdjuster.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize, basinhopping
from scipy.interpolate import UnivariateSpline
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Peaks.ElutionModels import egh_pdf

TAU_LIMIT_RATIO = 0.5   # relative to sigma
VERY_SMALL_VALUE = 1e-10
XRAY_SCALE = 1

class UvAdjuster:
    def __init__(self, xr_x, xr_y, xr_params):
        self.xr_x = xr_x
        self.xr_y = xr_y
        self.xr_params = xr_params

    def guess_initial_mapping_info(self, uv_x, uv_y, uv_spline):
        xr_x = self.xr_x
        ty = np.zeros(len(xr_x))
        cy_list = []
        for w, m, s, t in self.xr_params:
            cy = w * egh_pdf(xr_x, m, s, t)
            cy_list.append(cy)
            ty += cy

        max_ty = np.max(ty)
        xr_scale = np.max(self.xr_y)/max_ty
        max_uv_y = np.max(uv_y)
        uv_scale = max_uv_y/max_ty
        print("max_uv_y=", max_uv_y)
        print("uv_scale=", uv_scale)
        print("xr_scale=", xr_scale)

        if False:
            plt.push()
            fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(14,6))
            ax1.plot(uv_x, uv_y)
            ax1.plot(uv_x, uv_spline(uv_x))
            ax2.plot(xr_x, self.xr_y)
            for cy in cy_list:
                ax2.plot(xr_x, xr_scale*cy, ':')
            fig.tight_layout()
            debug_plot = plt.show()
            plt.pop()

        debug_plot = False

        def obj_func(p, return_scale=False):
            nonlocal debug_plot
            slope, intercept = p
            uv_ty = np.zeros(len(xr_x))
            uv_x = slope * xr_x + intercept
            y_ = uv_spline(uv_x)
            k = 0
            uv_cy_list = []
            for w, m, s, t in self.xr_params:
                uv_m = slope * m + intercept
                uv_s = slope * s
                uv_t = slope * t
                uv_cy = uv_scale * w * egh_pdf(uv_x, uv_m, uv_s, uv_t)
                uv_cy_list.append(uv_cy)
                uv_ty += uv_cy

            scale = max_uv_y/np.max(uv_ty)
            if return_scale:
                return scale

            fv = np.sum((scale*uv_ty - y_)**2)
            if debug_plot:
                print("p=", p)
                print("fv=", fv)
                # scale = xr_scale/uv_scale
                plt.push()
                fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(14,6))
                ax1.plot(uv_x, y_)
                ax2.plot(xr_x, self.xr_y)
                for cy in uv_cy_list:
                    ax1.plot(uv_x, scale*cy, ':')
                    ax2.plot(xr_x, xr_scale*cy, ':')
                ax1.plot(uv_x, scale*cy, ':')
                ax2.plot(xr_x, xr_scale*cy, ':')
                fig.tight_layout()
                debug_plot = plt.show()
                plt.pop()

            return fv
        init_params = uv_x[-1]/xr_x[-1], uv_x[0]
        bounds = [ (v*0.8, v*1.2) for v in init_params]
        res = minimize(obj_func, init_params, bounds=bounds)

        ret_scale = obj_func(res.x, return_scale=True)
        return ret_scale, res.x

    def obj_func(self, p):
        if self.fixed_mapping:
            w_ = p
            slope = self.slope
            intercept = self.intercept
        else:
            w_ = p[0:-2]
            slope, intercept = p[-2:]
        uv_ty = np.zeros(len(self.xr_x))
        uv_x = slope * self.xr_x + intercept
        k = 0
        uv_cy_list = []
        for _, m, s, t in self.xr_params:
            w = w_[k]
            uv_m = slope * m + intercept
            uv_s = slope * s
            uv_t = slope * t
            uv_cy = w * egh_pdf(uv_x, uv_m, uv_s, uv_t)
            uv_cy_list.append(uv_cy)
            uv_ty += uv_cy
            k += 1
        uv_y = self.uv_spline(uv_x)
        scale = np.max(uv_y)/np.max(uv_ty)
        fv = np.sum((scale*uv_ty - uv_y)**2)

        if self.debug_plot:
            print("fv=", fv)
            plt.push()
            fig, ax = plt.subplots()
            ax.plot(uv_x, uv_y, color='blue')
            for k, uv_cy in enumerate(uv_cy_list):
                ax.plot(uv_x, scale*uv_cy, ':')
            ax.plot(uv_x, scale*uv_ty, ':', color='red')
            fig.tight_layout()
            self.debug_plot = plt.show()
            plt.pop()

        return fv

    def fit(self, uv_x, uv_y, uv_spline, mapping_info=None):
        self.uv_x = uv_x
        self.uv_spline = uv_spline
        if mapping_info is None:
            scale, (slope, intercept) = self.guess_initial_mapping_params(uv_x, uv_y, uv_spline)
        else:
            scale, (slope, intercept) = mapping_info

        xr_params = self.xr_params.copy()
        self.fixed_mapping = False
        if self.fixed_mapping:
            self.slope = slope
            self.intercept = intercept
            init_params = xr_params[:,0]
        else:
            init_params = np.concatenate([xr_params[:,0], [slope], [intercept]])

        self.debug_plot = False

        if True:
            # method='Nelder-Mead'
            method = None
            self.result = res = minimize(self.obj_func, init_params, method=method)
        else:
            seed = np.random.randint(1000, 9999)
            self.result = res = basinhopping(self.obj_func, init_params, seed=seed)

        if self.fixed_mapping:
            self.mapping_params = np.array([slope, intercept])
            xr_params[:,0] = res.x
        else:
            self.mapping_params = res.x[-2:]
            xr_params[:,0] = res.x[0:xr_params.shape[0]]
        self.decomp_params = xr_params

    def get_fit_score(self):
        return np.log(self.obj_func(self.result.x)/len(self.xr_x))

    def plot_result(self, ax):
        slope, intercept = self.mapping_params
        uv_x = self.uv_x
        uv_ty = np.zeros(len(uv_x))
        k = 0
        uv_cy_list = []
        for w, m, s, t in self.decomp_params:
            uv_m = slope * m + intercept
            uv_s = slope * s
            uv_t = slope * t
            uv_cy = w * egh_pdf(uv_x, uv_m, uv_s, uv_t)
            uv_cy_list.append(uv_cy)
            uv_ty += uv_cy
            k += 1
        uv_y = self.uv_spline(uv_x)
        scale = np.max(uv_y)/np.max(uv_ty)

        for k, uv_cy in enumerate(uv_cy_list):
            ax.plot(uv_x, scale*uv_cy, ':', label='component-%d' % k)
        ax.plot(uv_x, scale*uv_ty, ':', color='red', label='total')

    def plot_mapped_uv(self, ax):
        slope, intercept = self.mapping_params
        x = self.xr_x
        uv_ty = np.zeros(len(x))
        k = 0
        uv_cy_list = []
        for w, m, s, t in self.decomp_params:
            uv_cy = w * egh_pdf(x, m, s, t)
            uv_cy_list.append(uv_cy)
            uv_ty += uv_cy
            k += 1

        uv_y = self.uv_spline(slope*x + intercept)
        y = np.zeros(len(x))
        for uv_cy, w_uv, w_xr in zip(uv_cy_list, self.decomp_params[:,0], self.xr_params[:,0]):
            y += w_xr/w_uv * uv_cy/uv_ty * uv_y

        scale = np.max(self.xr_y)/np.max(y)

        # for k, uv_cy in enumerate(uv_cy_list):
        #     ax.plot(x, scale*uv_cy, ':', label='component-%d' % k)
        ax.plot(x, scale*y, ':', color='blue', label='mapped uv')

def spike_demo(in_folder, sd=None, xr_params=None, seeds=None, mm_no=None):
    from molass_legacy.Batch.StandardProcedure import StandardProcedure
    from molass_legacy.Elution.CurveUtils import simple_plot

    if sd is None:
        sp = StandardProcedure()
        sp.load(in_folder, debug=False)
        sd = sp.get_corrected_sd()

    D, E, qv, xr_curve = sd.get_xr_data_separate_ly()
    U, wlvec, uv_curve = sd.conc_array, sd.lvector, sd.get_uv_curve()

    if xr_params is None:
        """
        # 6999, 7028
        xr_params = np.array([
            [  2.57017904, 151.13250995,  27.53165132,   7.36131756,  99.42063111],
            [ 14.46655509, 222.19960345,  22.67418396, -11.30840347,  41.30237997],
            [  2.27135188, 261.20370834,  14.33768941,   7.15074626,  38.81913523],
            [  1.13742087, 320.3231673,   31.64481979,  -1.19046219,  42.69479584],
            ])

        # 9475, 7228
        xr_params = np.array([
            [  2.14588686, 145.16722691,  17.95032162,  -5.48518373,  94.23021699],
            [  3.56763662, 183.06996861,  13.01316199,  -6.46241911,  57.03501218],
            [ 16.49286555, 220.10341326,  22.00469214,  10.9826917,   40.68061606],
            [  1.68483415, 312.14292219,  41.90521428, -10.98016292,  41.69664135],
            ])
        # 9749, 9287
        xr_params = np.array([
            [  2.55842938, 143.78127637,  22.34910787,  10.24668622,  99.2569971 ],
            [ 10.56334344, 204.95578495,  20.13302181,  -9.14981967,  47.5842871 ],
            [ 10.55376469, 232.60451753,  18.68320577,   9.33390851,  38.54086037],
            [  2.11340872, 303.27840531,  43.90177422,   3.14959293,  42.00333368],
            ])
        """
        mm_seed, rr_seed = 7063, 1769
        xr_params = np.array([
            [  2.72274354, 145.29605286,  17.47330807,  -7.60850013,  94.15078257],
            [  5.05797034, 183.27475203,  13.37163632,  -6.61815137,  57.59996813],
            [ 21.1954766,  220.11493197,  21.50917803,  10.7468006,   40.70055306],
            [  2.83886823, 297.77671412,  51.77808462,  24.13764017,  41.89812747],
            ])
    else:
        mm_seed, rr_seed = seeds

    def debug_plot(uv_curve, xr_curve, ua=None):
        dp = plt.push()
        if ua is None:
            fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(14,6))
        else:
            from DataUtils import get_in_folder
            fig, (ax1, ax2, ax3) = plt.subplots(ncols=3, figsize=(21,7))
            in_folder_ = get_in_folder(in_folder)
            fig.suptitle("Simultaneuos UV-scaling and Mapping for %s with seeds=(%d, %d)" % (in_folder_, mm_seed, rr_seed), fontsize=24)
            ax1.set_title("UV Scaled Decomposition", fontsize=20)
            ax2.set_title("Xray Decomposition", fontsize=20)
            ax3.set_title("Mapped UV Elution overlaid on Xray Elution", fontsize=20)

        simple_plot(ax1, uv_curve,  color='blue', legend=False)
        if ua is not None:
            ua.plot_result(ax1)
        ax1.legend(fontsize=16)

        simple_plot(ax2, xr_curve, color='orange', legend=False)

        x = xr_curve.x
        y = xr_curve.y

        ty = np.zeros(len(x))
        cy_list = []
        for w, m, s, t in xr_params[:,0:4]:
            cy = w * egh_pdf(x, m, s, t)
            cy_list.append(cy)
            ty += cy
        scale = xr_curve.max_y/np.max(ty)

        for k, cy in enumerate(cy_list):
            ax2.plot(x, scale*cy, ':', label='component-%d' % k)

        ax2.plot(x, scale*ty, ':', color='red', label='total')
        ax2.legend(fontsize=16)

        if ua is not None:
            simple_plot(ax3, xr_curve,  color='orange', legend=False)
            ua.plot_mapped_uv(ax3)
            ax3.legend(fontsize=16)

        fig.tight_layout()
        if ua is not None:
            fig.subplots_adjust(top=0.85)

        if mm_no is None:
            plt.show()
        else:
            path = get_figure_path(mm_no)
            dp.fig.canvas.draw()
            dp.fig.savefig(path)

        plt.pop()

    if False:
        debug_plot(uv_curve, xr_curve)

    x = xr_curve.x
    y = xr_curve.y
    uv_x = uv_curve.x
    uv_y = uv_curve.y
    uv_spline = uv_curve.spline
    ua = UvAdjuster(x, y, xr_params[:,0:4])
    mapping_info = ua.guess_initial_mapping_info(uv_x, uv_y, uv_spline)
    print("mapping_info=", mapping_info)
    ua.fit(uv_x, uv_y, uv_spline, mapping_info)
    fit_score = ua.get_fit_score()
    print("fit_score=", fit_score)

    if True:
        debug_plot(uv_curve, xr_curve, ua=ua)

    return fit_score

def get_figure_path(mm_no):
    import os
    from molass_legacy._MOLASS.SerialSettings import get_setting
    sub_no = 0
    while True:
        sub_no += 1
        path  = os.path.join(get_setting("temp_folder"), "fig-%03d-%d.jpg" % (mm_no, sub_no))
        if os.path.exists(path):
            continue
        else:
            break
    return path

def compute_all(in_folder, log_file=None):
    import logging
    import re
    from molass_legacy.Batch.StandardProcedure import StandardProcedure
    from molass_legacy.Elution.CurveUtils import simple_plot
    from NumpyArrayUtils import from_space_separated_list_string

    sp = StandardProcedure()
    sp.load(in_folder, debug=False)
    sd = sp.get_corrected_sd()

    logger = logging.getLogger()
    parsing_params = False
    parsing_start = False
    params_key = "params="
    mixture_re = re.compile(r"Mixture\((\d+)\)")
    refiner_seeds_re = re.compile(r"Refiner\((\d+), (\d+)\)")
    trucate_re = re.compile(r"^(.+\]\]), func_value=(.+)")

    seeds_list = []
    params_list = []
    fv_list = []
    score_list = []
    classify_key_list = []
    mm_no = -1
    if log_file is None:
        # log_file = r"C:\Users\takahashi\Dropbox\_MOLASS\meeting-docs\20210426\rgg-figs\20190529_4\test.log"
        # log_file = r"D:\PyTools\pytools-2_2_4_develop\test\temp\test.log"
        log_file = r"D:\TODO\20210426\rgg-figs\test.log"
    fh = open(log_file)
    params_count = 0
    for k, line in enumerate(fh.readlines()):
        if params_count >= 9:
            # break
            pass

        if line.find("Mixture") > 0:
            m = mixture_re.search(line)
            if m:
                mm_no += 1

        if line.find("Refiner") > 0:
            parsing_params = True
            parsing_start = True
            m = refiner_seeds_re.search(line)
            if m:
                seeds = [int(m.group(n)) for n in [1,2]]
            else:
                assert False
        if parsing_params:
            # print([k], line)
            if parsing_start:
                parsing_start = False
                pos = line.find(params_key) + len(params_key)
                params_str = line[pos:]
                params_count += 1
            else:
                if line.find("]]") > 0:
                    m = trucate_re.match(line)
                    if m:
                        params_str += m.group(1)
                        fv = float(m.group(2))
                    else:
                        assert False

                    if np.isfinite(fv):
                        fv_list.append(fv)
                        parsing_params = False
                        params = from_space_separated_list_string(params_str)
                        logger.info("seeds=%s, params={%s}", str(seeds), str(params))
                        fit_score = spike_demo(in_folder, sd=sd, xr_params=params, seeds=seeds, mm_no=mm_no)
                        logger.info("fit_score=%g", fit_score)
                        seeds_list.append(seeds)
                        params_list.append(params)
                        score_list.append(fit_score)
                        # classify_key_list.append(np.concatenate([params[:,1], [fv]]))
                        classify_key_list.append(np.concatenate([[np.std(params[:,1])/10], [fv + fit_score]]))
                else:
                    params_str += line
    fh.close()

    classify_key_array = np.array(classify_key_list)

    print("classify_key_array.shape=", classify_key_array.shape)
    np.savetxt("classify_key_array.txt", classify_key_array)

    # X = classify_key_array[:,0:2]
    X = classify_key_array
    if True:
        from sklearn.cluster import KMeans
        kmeans = KMeans(n_clusters=5, random_state=0).fit(X)
        labels = kmeans.labels_
    else:
        from sklearn.mixture import GaussianMixture
        gmm = GaussianMixture(n_components=5).fit(X)
        labels = gmm.predict(X)
    if True:
        plt.push()
        fig, ax = plt.subplots()
        ax.scatter(X[:, 0], X[:, 1], c=labels, s=40, cmap='viridis')
        plt.show()
        plt.pop()

    fh = open("classification.csv", "w")
    for seeds, label, fv, score in zip(seeds_list, labels, fv_list, score_list):
        print((seeds, label, fv, score))
        fh.write(','.join([str(v) for v in [*seeds, label, fv, score]])+'\n')

    fh.close()



"""
    XrayLrfConcProof.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np
from bisect import bisect_right
import molass_legacy.KekLib.DebugPlot as plt
from SimTools.HardSphere import get_model_data
from SimTools.EoiiPlotUtils import plot_eoii

def demo():
    qv = np.linspace(0.005, 0.4, 200)
    jv = np.arange(300)
    rg = 35
    h = 1
    mu = 150
    sigma = 20
    tau = 0
    K = 0.5
    D, P, C = get_model_data(qv, jv, rg, h, mu, sigma, tau, K=K)

    title = "Eoii Algorithm simply using Aparent Data"

    i = bisect_right(qv, 0.02)

    with plt.Dp():
        fig, axes = plt.subplots(nrows=2, ncols=3, figsize=(18,10))
        fig.suptitle(title, fontsize=20)
        ax1, ax2, ax3 = axes[0,:]
        ax1.set_title("Elution View", fontsize=16)
        ax2.set_title("Scattering View (Linear)", fontsize=16)
        ax3.set_title("Scattering View (Log)", fontsize=16)
        plot_eoii(qv, jv, D, P, C, axes[0,:])

        A = D.copy()
        cy_init = A[i,:]
        cy = cy_init
        cy2 = cy**2
        C_ = np.array([cy, cy2])
        C_inv = np.linalg.pinv(C_)
        P_ = D @ C_inv

        plot_eoii(qv, jv, D, P_, C_, axes[1,:])

        fig.tight_layout()
        plt.show()

def proof():
    from LRF.XrayLrfConc import estimate_true_concentration

    qv = np.linspace(0.005, 0.4, 200)
    jv = np.arange(300)
    rg = 35
    h = 1
    mu = 150
    sigma = 20
    tau = 0
    K = 0.5
    D, P, C = get_model_data(qv, jv, rg, h, mu, sigma, tau, K=K)

    title = "Eoii Algorithm simply using Aparent Data"

    with plt.Dp():
        fig, axes = plt.subplots(nrows=2, ncols=3, figsize=(18,10))
        fig.suptitle(title, fontsize=20)
        ax1, ax2, ax3 = axes[0,:]
        ax1.set_title("Elution View", fontsize=16)
        ax2.set_title("Scattering View (Linear)", fontsize=16)
        ax3.set_title("Scattering View (Log)", fontsize=16)
        plot_eoii(qv, jv, D, P, C, axes[0,:])

        C_ = estimate_true_concentration(qv, D)
        C_inv = np.linalg.pinv(C_)
        P_ = D @ C_inv

        plot_eoii(qv, jv, D, P_, C_, axes[1,:])

        fig.tight_layout()
        plt.show()



# coding: utf-8
"""
    Pomegranate.Spike.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
from time import time
import asyncio
import numpy as np
# from pomegranate import *
from sklearn.neighbors import KernelDensity
from molass_legacy.KekLib.SciPyCookbook import smooth
from molass_legacy.Peaks.ElutionModels import egh
import molass_legacy.KekLib.DebugPlot as plt
from Pomegranate.EghMixtureModel import EghMixtureModel
from Prob.GaussianMixture import hist_to_source

def normal_pdf(x, mu, sigma):
    return np.exp(-1/2*((x - mu)/sigma)**2)/(sigma*np.sqrt(np.pi))

def distributions_demo():
    x = np.arange(500)

    plt.push()
    fig, ax = plt.subplots()

    d = EghDistribution(200, 30, 5)
    ax.plot(x, d.probability(x))

    fig.tight_layout()
    plt.show()
    plt.pop()

def spike_demo():
    x = np.arange(500)
    y = np.zeros(len(x))
    for params in [
            # [0.6, 200, 40, 10],
            [0.6, 200, 30, 5],
            [1.0, 300, 30, 10],
            ]:
        y += egh(x, *params)

    spike_demo_impl([(x, y)], num_components=2)

def spike_demo_impl(xy_list, num_components):
    for x, y in xy_list:
        sy = smooth(y)
        X = hist_to_source(x, sy)[:,np.newaxis] 
        kde = KernelDensity(kernel='gaussian', bandwidth=20).fit(X)
        # print(dir(kde))
        log_dens = kde.score_samples(x[:,np.newaxis])
        fy = np.exp(log_dens)
        scale = np.max(y)/np.max(fy)

        plt.push()
        fig, ax = plt.subplots()
        ax.plot(x, y)
        ax.plot(x, fy*scale)
        plt.show()
        plt.pop()

def spike_demo_real(in_folder, num_components=3, correction=False, use_peaks=False):
    from molass_legacy.Batch.StandardProcedure import StandardProcedure
    sp = StandardProcedure()
    sp.load(in_folder, debug=False)
    if correction:
        sd = sp.get_corrected_sd()
    else:
        sd = sp.get_sd()
    uv_curve = sd.get_uv_curve()
    xr_curve = sd.get_xray_curve()
    xy_list = [(curve.x, curve.y) for curve in [uv_curve, xr_curve]]
    spike_demo_impl(xy_list, num_components=3)

async def get_content(n):
    await asyncio.sleep(3)
    return n + 1

async def f(n):
    content = await get_content(n)
    return content

def coroutine_spike():
    loop = asyncio.get_event_loop()
    v = loop.run_until_complete(f(1))
    print(v)



# coding: utf-8
"""
    ファイル名：   MinimalTiff.py

    処理内容：

        ヘッダ情報を変えない（または必要最小の書き換え）で、
        データのみを書き換えるためのモジュール

    Copyright (c) 2016, Masatsuyo Takahashi, KEK-PF
"""
import sys
import numpy    as np

DEBUG = 0

class IfdEntry:
    def __init__( self, buffer ):
        assert( len( buffer ) == 12 )
        self.tag    = '{:04x}'.format( int.from_bytes( buffer[0:2], 'little'  ) )
        self.type   = int.from_bytes( buffer[2:4], 'little' )
        self.value1 = int.from_bytes( buffer[4:8], 'little' )
        self.value2 = int.from_bytes( buffer[8:12], 'little' )

    def __str__( self ):
        return '[' + ','.join( [ self.tag, str( self.type ), str( self.value1 ), str( self.value2 ) ] ) + ']'

class IFD:
    def __init__( self, block, num_entries ):
        self.entry_dict = {}
        for i in range( num_entries ):
            start = i * 12
            entry = IfdEntry( block[start:start+12] )
            self.entry_dict[ entry.tag ] = [ entry.type, entry.value1, entry.value2 ]
            if DEBUG:
                print( '  %s : %s' % ( entry.tag, str( [ entry.type, entry.value1, entry.value2 ] ) ) )

    def __str__( self ):
        return str( self.entry_dict )

class MinimalTiff:
    def __init__( self, path, header_info_only=False ):

        fh = open( path, 'rb' )
        header_length = 8
        self.header = fh.read( header_length )
        assert( self.header[0:4] == b'\x49\x49\x2A\x00' )

        offset = int.from_bytes( self.header[4:6], 'little' )
        # print( 'offset=x%x (%d)' % ( offset, offset ) )

        self.pre_ifd_texts = []

        to_read_bytes = offset - header_length
        block1 = b''
        if to_read_bytes > 0:
            block1 = fh.read( to_read_bytes )

            nonzero_bytes = ''
            for byte in block1[0:-2]:
                if byte == 0:
                    if len( nonzero_bytes ) > 0:
                        # print( nonzero_bytes )
                        self.pre_ifd_texts.append( nonzero_bytes )
                        nonzero_bytes = ''
                else:
                    # print( chr( byte ) )
                    nonzero_bytes += chr( byte )

        self.header += block1

        word = fh.read( 2 )
        self.header += word

        num_entries = int.from_bytes( word, 'little' )
        if DEBUG:
            print( 'num_entries=', num_entries )

        to_read_bytes = num_entries * 12
        block2 = fh.read( to_read_bytes )
        self.header += block2

        """
        thanks to: http://www.awaresystems.be/imaging/tiff/tifftags.html
        """
        self.ifd = ifd = IFD( block2, num_entries )

        # Baseline Tags
        self.cols   = ifd.entry_dict[ '0100' ][2]   # ImageWidth
        self.rows   = ifd.entry_dict[ '0101' ][2]   # ImageLength
        num_bits    = ifd.entry_dict[ '0102' ][2]   # BitsPerSample
        compression = ifd.entry_dict[ '0103' ][2]   # Compression
        data_offset = ifd.entry_dict[ '0111' ][2]   # StripOffsets
        data_size   = ifd.entry_dict[ '0117' ][2]   # StripByteCounts

        # Extension Tags
        tag_0153    = ifd.entry_dict.get( '0153' )  # SampleFormat
        data_type   = tag_0153[2] if tag_0153 else None

        to_read_bytes = data_offset - len( self.header )

        if DEBUG:
            print( 'data offset=', data_offset  )
            print( 'data size=', data_size )
            print( 'data rows=', self.rows   )
            print( 'data cols=', self.cols  )
            print( 'data type=', data_type  )
            print( 'to_read_bytes=', to_read_bytes )

        if compression != 1:
            sys.exit( "Sorry, this module can't read compressed files!" )

        block3 = fh.read( to_read_bytes )
        self.header += block3

        data_buffer = fh.read()
        # print( 'len(data_buffer)=', len( data_buffer ) )
        assert( len( data_buffer ) == data_size )

        num_bytes = num_bits // 8

        dtype_ = 'i%d' % num_bytes
        if data_type == 2:
            pass
        elif data_type == 3:
            dtype_ = 'f%d' % num_bytes
        else:
            pass
        self.data = np.frombuffer( data_buffer, dtype=dtype_ )
        self.data.shape = ( self.rows, self.cols )
        self.dtype = self.data.dtype

        fh.close()

    def save( self, path ):
        if self.data.shape != ( self.rows, self.cols ):
            sys.exit( "You can't change the shape of the data." )

        if self.data.dtype != self.dtype:
            sys.exit( "You can't change the dtype of the data." )

        fh = open( path, 'wb' )
        fh.write( self.header )
        fh.write( self.data.tobytes() )
        fh.close()



# coding: utf-8
"""
    PlotUtils.py

    Copyright (c) 2016-2020, Masatsuyo Takahashi, KEK-PF
"""
import numpy                as np

def get_wider_range( vmin, vmax, ratio ):
    vmin_   = vmin * ( 1 + ratio ) + vmax * ( -ratio )
    vmax_   = vmin * ( -ratio ) + vmax * ( 1 + ratio )
    return vmin_, vmax_

def convert_to_the_level( yval, ymin, ymax, posf, post, vmin=None, vmax=None ):
    # print( 'convert_to_the_level: vmin, vmax=', vmin, vmax )
    if vmin is None:
        vmin = np.min( yval )
    if vmax is None:
        vmax = np.max( yval )
    ymin_ = ymin * ( 1 - posf ) + ymax * posf
    ymax_ = ymin * ( 1 - post ) + ymax * post
    ratio = ( ymax_ - ymin_ ) / ( vmax - vmin )
    return ( yval - vmin ) * ratio + ymin_

def plot_line( ax, x, y, rec, color, label=None, plot_qrg=False ):
    Rg, a, b, f, t = rec

    a_ = - ( Rg**2 / 3 )
    y0 = b + a_ * x[f]
    y1 = b + a_ * x[t]

    print( 'plot_line: Rg=', Rg, '; a=', a )
    print( [ x[f], x[t] ], [ y0, y1 ] )

    ax.plot( [ x[f], x[t] ], [ y0, y1 ], marker='o', color=color, label=label )

"""
    https://stackoverflow.com/questions/10481990/matplotlib-axis-with-two-scales-shared-origin
"""
def align_yaxis_np(ax1, ax2):
    """Align zeros of the two axes, zooming them out by same ratio"""
    axes = np.array([ax1, ax2])
    extrema = np.array([ax.get_ylim() for ax in axes])
    tops = extrema[:,1] / (extrema[:,1] - extrema[:,0])
    # Ensure that plots (intervals) are ordered bottom to top:
    if tops[0] > tops[1]:
        axes, extrema, tops = [a[::-1] for a in (axes, extrema, tops)]

    # How much would the plot overflow if we kept current zoom levels?
    tot_span = tops[1] + 1 - tops[0]

    extrema[0,1] = extrema[0,0] + tot_span * (extrema[0,1] - extrema[0,0])
    extrema[1,0] = extrema[1,1] + tot_span * (extrema[1,0] - extrema[1,1])
    [axes[i].set_ylim(*extrema[i]) for i in range(2)]



# coding: utf-8
"""
    Baseline.MpiAlgebra.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np
from bisect import bisect_right
from matplotlib.gridspec import GridSpec
from mpl_toolkits.mplot3d import Axes3D
from matplotlib.patches import Polygon
from MeasuredData import MeasuredData
import molass_legacy.KekLib.DebugPlot as plt
from MatrixData import simple_plot_3d
from .Baseline import compute_baseline
from molass_legacy.SerialAnalyzer.ElutionCurve import ElutionCurve
from CurveDecomposer import decompose
from molass_legacy.Models.ElutionCurveModels import egh
from DataModels import GuinierPorod
from.PatchUtils import make_area_points

TITLE_FONTSIZE = 16

def drifted_elution(x):
    y1 = egh(x, 1, 100, 25, 5)
    y2 = egh(x, 0.8, 180, 20, 10)
    y_ = y1 + y2
    d = np.cumsum(y_*0.001)
    return np.array([y1, y2, d])

def demo(in_folder=None, split=True):
    q = np.arange(600)*0.001
    i = bisect_right(q, 0.02)
    x = np.arange(300)
    C = drifted_elution(x)

    p_list = []
    for k, params in enumerate([(50, 3, 100), (30, 2, 180), (10, 1, 299) ]):
        rg, d, j = params
        gp = GuinierPorod(1, rg, d)
        y = gp(q)
        p_list.append(y*C[k,j]/y[i])
    P = np.array(p_list).T
    M = np.dot(P, C)

    fig = plt.figure(figsize=(21,11))
    gs = GridSpec(2,3)
    ax00 = fig.add_subplot(gs[0,0], projection='3d')
    ax01 = fig.add_subplot(gs[0,1])
    ax02 = fig.add_subplot(gs[0,2])
    ax11 = fig.add_subplot(gs[1,1])
    ax12 = fig.add_subplot(gs[1,2])

    ax00.set_title("3D Data View", fontsize=TITLE_FONTSIZE, y=1.09)
    ax01.set_title("Component Elution Curves", fontsize=TITLE_FONTSIZE)
    ax02.set_title("Component Scattering Curves", fontsize=TITLE_FONTSIZE)

    if split:
        paren1 = "(split)"
        paren2 = "(overlaid with split ones)"
    else:
        paren1 = "(summed)"
        paren2 = "(overlaid with summed one)"
    ax11.set_title("Component Elution Curves " + paren1, fontsize=TITLE_FONTSIZE)
    ax12.set_title("Component Scattering Curves " + paren2, fontsize=TITLE_FONTSIZE)

    simple_plot_3d(ax00, M, x=q)
    qi = np.ones(len(x))*q[i]
    y = x
    z = M[i,:]
    ax00.plot(qi, y, z, color='orange')

    ty = np.zeros(len(x))

    for k, c in enumerate(C):
        ty += c
        for ax in [ax01, ax11]:
            ax.plot(x, c, ':', label='C[%d,:]' % k)

    for ax in [ax01, ax11]:
        ax.plot(x, ty, color='orange', label='total')

    for k, y in enumerate(p_list):
        for ax in [ax02, ax12]:
            ax.plot(q, y, ':', label='P[:,%d]' % k)

    Pmpi = np.dot(M, np.linalg.pinv(C))

    if False:
        for y in Pmpi.T:
            ax12.plot(q, y)

    if split:
        w = np.random.uniform(0,1,300)
        ca = C[0,:]*w
        cb = C[0,:]*(1-w)
        points = make_area_points(x, C[0,:], ca)
        area = Polygon(points, alpha=0.3, color='cyan')
        ax11.add_patch(area)
        points = make_area_points(x, ca)
        area = Polygon(points, alpha=0.3, color='pink')
        ax11.add_patch(area)
        w = np.random.uniform(0,1,300)
        cc = C[1,:]*w
        cd = C[1,:]*(1-w)
        points = make_area_points(x, C[1,:], cc)
        area = Polygon(points, alpha=0.3, color='yellow')
        ax11.add_patch(area)
        points = make_area_points(x, cc)
        area = Polygon(points, alpha=0.3, color='green')
        ax11.add_patch(area)
        C_ = np.array([ca, cb, cc, cd, C[2,:]])
        P_ = np.dot(M, np.linalg.pinv(C_))

        for k, y in enumerate(P_.T):
            if k < 4:
                ul = '_upper' if k%2 == 0 else '_lower'
            else:
                ul = ''
            ax12.plot(q, y, '-', markersize=1, label='P_[:,%d]%s' % (k//2, ul))

    else:
        cs = C[0,:]+C[1,:]
        ax11.plot(x, cs, label='C[0,:]+C[1,:]')

        C_ = np.array([cs, C[2,:]])
        P_ = np.dot(M, np.linalg.pinv(C_))

        for k, y in enumerate(P_.T):
            ax12.plot(q, y, '-', markersize=1, label='P_[:,%d]' % k)

    for ax in [ax01, ax02, ax11, ax12]:
        ax.legend()

    fig.tight_layout()
    plt.show()



"""
    FitRecord.py

    Copyright (c) 2018-2024, SAXS Team, KEK-PF
"""
import copy

USE_MODEL_TO_GET_RANGES = True
MIN_SEPARATE_RANGE_LENGTH = 10
MIN_SEPARATE_AREA_RATIO = 0.1
IGNORABLE_AREA_RATIO = 0.03

class FitRecord:
    def __init__(self, kno, evaluator, chisqr_n, peak, major_pno=None, area=None):
        self.kno    = kno       # negative for smaller peaks
        self.evaluator = evaluator
        self.chisqr_n = chisqr_n
        self.peak = peak
        self.major_pno = major_pno
        self.area = area
        self.item_list = [self.kno, self.evaluator, self.chisqr_n, self.peak, self.major_pno, self.area]

    def set_evaluator(self, evaluator):
        self.evaluator = evaluator
        self.item_list[1] = evaluator

    def get_params(self):
        return self.evaluator.get_all_param_values()

    def scale_to(self, other, conc_factor):
        new_evaluator = copy.deepcopy(self.evaluator)   # without this copy, self.evaluator would be changed
        new_rec = FitRecord(self.kno, new_evaluator, self.chisqr_n, self.peak, self.major_pno)
        other_h = other.evaluator.get_param_value(0)
        new_rec.evaluator.update_param(0, other_h*conc_factor)
        if True:
            this_h_old = self.evaluator.get_param_value(0)
            this_h_new = new_rec.evaluator.get_param_value(0)
            print('scale_to:', this_h_old, other_h, this_h_new)
        return new_rec

    def __getitem__(self, index):
        return self.item_list[index]

    def __repr__(self):
        return "<FitRecord %s>" % str(self.item_list)

    def __lt__(self, other):
        # neccesary in case when sorted( temp_pair_list ) fails
        return self.evaluator.get_param_value(1) < other.evaluator.get_param_value(1)

    def get_range_list(self, x, debug=True):
        # print('get_range_list: self=', self)

        rec_id = self.kno
        peak = self.peak

        if debug:
            print('get_range_list: peak.area_prop=', peak.area_prop)

        if USE_MODEL_TO_GET_RANGES:
            model_evaluator = self[1]
            foot_L, top_x, foot_R = model_evaluator.get_range_params(x)
            if (rec_id >= 0 and foot_R - foot_L >= MIN_SEPARATE_RANGE_LENGTH
                and (peak.area_prop is None or peak.area_prop > MIN_SEPARATE_AREA_RATIO)):
                top_x_ = int(round(top_x))
                range_list = [ [ foot_L, top_x_ ], [ top_x_, foot_R ] ]
            else:
                range_list = [ [ foot_L, foot_R ] ]
        else:
            if rec_id >= 0:
                range_list = [ [ peak.sigma_points[0], peak.top_x_m ], [ peak.top_x_m, peak.sigma_points[1] ] ]
            else:
                range_list = [ [ peak.sigma_points[0], peak.sigma_points[1] ] ]

        return range_list

def sort_fit_recs(recs):
    return sorted(recs, key=lambda p:p.peak.top_x)



# coding: utf-8
"""
    SaxsDetector.py

    Copyright (c) 2019, SAXS Team, KEK-PF
"""
import os
import numpy as np
from scipy import ndimage, interpolate
import matplotlib.pyplot as plt
import matplotlib.cm as cm
from .ReciprocalData import ReciprocalData

USE_ZERNIKE_COEF = False

class SaxsDetector:
    def __init__(self, fig, ax, data, q, in_y):
        ax.set_title("Detector Image", fontsize=20)
        ax.set_axis_off()

        if USE_ZERNIKE_COEF:
            from .OurOptics import ZernikeCoefficient
            Z = ZernikeCoefficient(Z11=1)
            Z.zernikemap(fig=fig, ax=ax, cmap=cm.plasma, label=False)
            return

        rdata = ReciprocalData(data.shape)
        F = rdata.get_reciprocal(data)
        self.curve_y = y = rdata.get_scattering_curve(q, F)
        self.draw_detector_image(fig, ax, q, y, in_y)

    def compute_zernike_coefficients(self, data):
        # TODO
        pass


    def draw_detector_image(self, fig, ax, q, y, in_y):
        assert len(q) == len(y)
        n = len(y)
        qmax = q[-1]
        theta = np.linspace(0, 2*np.pi, 400)
        rho = np.linspace(0, qmax, 400)
        u,r = np.meshgrid(theta,rho)
        X = r*np.cos(u)
        Y = r*np.sin(u)
        # r -> y
        scale = 1 if in_y is None else in_y[0]/y[0]
        interp = interpolate.interp1d(q, y*scale, kind='cubic', fill_value="extrapolate")
        Z = interp(r)
        im = ax.pcolormesh(X, Y, Z, cmap=cm.plasma)
        fig.colorbar(im, ax=ax)
        ax.set_aspect('equal', 'datalim')



"""
    Basseline.BoBaseline.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Baseline.Constants import SLOPE_SCALE

USE_BAYESIAN_OPTIMIZION = True

def bo_baseline_trial_impl(caller):
    from importlib import reload
    import UV.UvPreRecog
    reload(UV.UvPreRecog)
    from molass_legacy.UV.UvPreRecog import UvPreRecog
    from molass_legacy.Trimming.PreliminaryRecognition import PreliminaryRecognition
    print('bo_baseline_trial_impl')

    sd = caller.serial_data
    pre_recog = PreliminaryRecognition(sd)

    upr = UvPreRecog(sd, pre_recog, debug=True)

    base_curve = upr.base_curve
    base_params = upr.init_params.copy()
    base_params[4:6] /= SLOPE_SCALE
    uv_curve = sd.get_uv_curve()
    x = uv_curve.x
    y = uv_curve.y
    ty = None
    by = base_curve(x, base_params, ty)

    print("baseline_type=", base_curve.baseline_type)
 
    def plot_baseline(title, plot_basepart=True):
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.set_title(title)
            ax.plot(x, y)
            if plot_basepart:
                ax.plot(tbx, tby)
            ax.plot(x, by, color='red')
            fig.tight_layout()
            plt.show()

    tbx, tby = base_curve.base_xy
    plot_baseline("bo_baseline_trial_impl: before minimize")
    return

    def objective_func(params):
        y_ = base_curve(tbx, params, ty)
        return np.sum((y_ - tby)**2)

    if USE_BAYESIAN_OPTIMIZION:
        import logging
        from ultranest import ReactiveNestedSampler
        from molass_legacy.Optimizer.OptimizerUtils import OptimizerResult
        allow = 10
        upper = base_params - allow
        lower = base_params + allow

        def my_prior_transform(cube):
            params = cube * (upper - lower) + lower
            return params
        def my_likelihood(params):
            return -objective_func(params)

        num_params = len(base_params)
        param_names = ["p%02d" % i for i in range(num_params)]
        sampler = ReactiveNestedSampler(param_names, my_likelihood, my_prior_transform)
        sampler.logger.setLevel(logging.INFO)       # to suppress debug log
        result = sampler.run(min_num_live_points=100, max_ncalls=20000)
        opt_params = np.array(result['maximum_likelihood']['point'])
        res = OptimizerResult(x=opt_params)
    else:
        from scipy.optimize import basinhopping
        minimizer_kwargs = dict(method='Nelder-Mead')
        res = basinhopping(objective_func, base_params, minimizer_kwargs=minimizer_kwargs)

    by = base_curve(x, res.x, ty)
    plot_baseline("bo_baseline_trial_impl: after minimize")



# coding: utf-8
"""

    LPM_Animation.py

    Copyright (c) 2018, Masatsuyo Takahashi, KEK-PF

"""
import copy
import numpy            as np
import matplotlib.pyplot as plt
import matplotlib.animation as animation
from lmfit              import minimize, Parameters
from scipy              import stats
from molass_legacy._MOLASS.SerialSettings     import get_setting

PERCENTILE_FIRST        = 25
PERCENTILE_SECOND       = 25
PERCENTILE_FINAL        = 10
VERY_SMALL_SLOPE_RATIO  = 0.01
CONVERGENCE_RATIO       = 0.1

def animation_example():
    fig = plt.figure()
    x = np.arange(0, 10, 0.1)

    ims = []
    for a in range(50):
        y = np.sin(x - a)
        line, = plt.plot(x, y, "r")
        ims.append([line])

    ani = animation.ArtistAnimation(fig, ims)
    # ani.save('anim.gif', writer="imagemagick")
    # ani.save('anim.mp4', writer="ffmpeg")
    plt.show()

def animation_example2():
    fig = plt.figure()
    x = np.arange(0, 10, 0.1)

    ims = []
    for a in range(50):
        y = np.sin(x - a)
        im = plt.plot(x, y, "b")
        ims.append(im)

    ani = animation.ArtistAnimation(fig, ims)
    ani.save('sample.gif', writer='imagemagick')

class LPM_AnimationDialog:
    pass


class ScatteringBaseline:
    def __init__( self, y, height=None, anim=False ):
        self.y = y
        self.x = x = np.arange( len(self.y) )
        if anim:
            sbl = ScatteringBaseline(y, anim=False)
            sbl.solve(p_final=7)
            n = sbl.npp
            self.fig, self.axes = plt.subplots(ncols=2, figsize=(16,7))
            self.fig.suptitle("Low Percentile Method Animation", fontsize=30)
            for ax in self.axes:
                ax.plot( y, color='gray' )

            ax1, ax2 = self.axes
            xmin, xmax = ax1.get_xlim()
            ymin, ymax = ax1.get_ylim()
            ax1.set_xlim(xmin, xmax)
            ax1.set_ylim(ymin, ymax)

            width = xmax - xmin
            height = ymax - ymin
            scale = 0.1

            fx, fy = x[n], y[n]
            zxmin = fx - width*scale/2
            zxmax = fx + width*scale/2
            zymin = fy - height*scale/2
            zymax = fy + height*scale/2

            ax2.set_xlim(zxmin, zxmax)
            ax2.set_ylim(zymin, zymax)

            self.fig.tight_layout()
            self.fig.subplots_adjust(top=0.9)
            self.anim_data_lists = [[],[]]

        self.anim = anim
        if height is None:
            height  = np.max( y ) - np.min( y )

        self.very_small_sploe = height / len(y) * VERY_SMALL_SLOPE_RATIO
        print( 'self.very_small_sploe=', self.very_small_sploe )

    def solve( self, p_final=PERCENTILE_FINAL, max_iter_num=10, no_shift=False ):
        xpp, ypp, slope, intercept, ppp = self.get_low_percentile_params( self.x, self.y, max_iter_num )
        print( 'ppp=', ppp )
        y_  = self.y - slope * self.x
        p   = np.percentile( y_, p_final )
        p_  = np.where( y_ <= p )[0]
        m   = np.argmax( y_[p_] )
        n   = p_[m]

        A   = slope
        if no_shift:
            B   = intercept
        else:
            B   = self.y[n] - slope*n

        if self.anim:
            for k, ax in enumerate(self.axes):
                anim_data = self.anim_data_lists[k]
                xmin, xmax = ax.get_xlim()
                ymin, ymax = ax.get_ylim()
                tx  = (xmin + xmax)/2
                ty  = (ymin + ymax)/2
                text = ax.text(tx, ty, "Shift", alpha=0.1, fontsize=200, ha="center", va="center")

                def compute_line( a, b ):
                    xp = []
                    yp = []
                    for x_ in self.x[ [0, -1] ]:
                        xp.append( x_ )
                        yp.append( a * x_ + b )
                    return xp, yp

                lpm_points = anim_data[-1][-1]
                point1, = ax.plot(p_, self.y[p_], 'o', color="yellow")
                point2, = ax.plot(n, self.y[n], 'o', color="red", markersize=10)

                xp, yp = compute_line( A, intercept )
                line, = ax.plot(xp, yp, "r")

                anim_data.append( [text, line, lpm_points, point1, point2] )

                if False:
                    xp, yp = compute_line( A, ( intercept + B )/2 )
                    line, = ax.plot(xp, yp, "r")
                    anim_data.append( [text, line, lpm_points, point1, point2] )

                xp, yp = compute_line( A, B )
                line, = ax.plot(xp, yp, "r")

                anim_data.append( [text, line, lpm_points, point1, point2] )

        self.xpp    = xpp
        self.ypp    = ypp
        self.npp    = n
        print('Base point: ', n)
        self.params = ( A, B )
        return A, B

    def is_alternating( self, slope_list ):
        slope_array = np.array(slope_list)
        half_iter_num = len(slope_list)
        average_slope = np.average(slope_array)
        slope_devs = slope_array - average_slope
        total_stdev = np.std(slope_array)
        upper_ratio = np.std(slope_array[slope_devs > 0])/total_stdev
        lower_ratio = np.std(slope_array[slope_devs < 0])/total_stdev
        # print('ratios=', upper_ratio, lower_ratio)
        alternating = (upper_ratio + lower_ratio) < ALTERNATING_LIMIT_RATIO
        # this is True for 20181203 (only as of Jan. 2019)
        return alternating, average_slope

    def get_low_percentile_params( self, x, y, max_iter_num ):
        y_  = y
        ppp = np.percentile( y_, [ PERCENTILE_FIRST ] )
        last_slope  = None
        init_diff   = None
        alternating = False
        last_xpp_avg = None
        half_iter_num = max_iter_num//2
        slope_list = []
        x_center = np.average(x)

        for i in range(max_iter_num):
            altered = False
            if i == half_iter_num:
                alternating, average_slope = self.is_alternating( slope_list )
                if alternating:
                    if self.logger is not None:
                        self.logger.warning( 'alternating state detected in LPM' )

            if alternating:
                if i > half_iter_num:
                    average_slope = np.average(slope_list[-half_iter_num:])
                slope = average_slope
            else:
                xpp = np.where( y_ <= ppp[0] )[0]
                ypp = y[xpp]
                slope, intercept, r_value, p_value, std_err = stats.linregress( xpp, ypp )

                xpp_avg = np.average(xpp)
                if last_xpp_avg is not None:
                    altered = (xpp_avg - x_center)*(last_xpp_avg - x_center) < 0
                    print([i], last_xpp_avg, xpp_avg, altered)

                if altered:
                    mid_slope = (slope + last_slope)/2
                    print([i], "slope should be modified from %g to %g" % (slope, mid_slope))
                    # slope = mid_slope

                last_xpp_avg = xpp_avg

            if self.anim:
                print( 'frame', i, ppp )

                for k, ax in enumerate(self.axes):
                    anim_data = self.anim_data_lists[k]
                    xmin, xmax = ax.get_xlim()
                    ymin, ymax = ax.get_ylim()
                    tx  = (xmin + xmax)/2
                    ty  = (ymin + ymax)/2
                    text = ax.text(tx, ty, str(i), alpha=0.1, fontsize=360, ha="center", va="center")

                    xp = []
                    yp = []
                    for x_ in self.x[ [0, -1] ]:
                        xp.append( x_ )
                        yp.append( slope * x_ + intercept )
                    line, = ax.plot(xp, yp, "r")

                    lpm_points, = ax.plot( xpp, ypp, 'o', color='orange' )

                    anim_data.append( [text, line, lpm_points] )

            y_ = y - slope*x
            ppp = np.percentile( y_, [ PERCENTILE_SECOND ] )

            print( 'slope=', slope )

            if last_slope is not None:
                diff = abs( slope - last_slope )
                if init_diff is None:
                    init_diff = diff
                if ( diff  < self.very_small_sploe
                    or diff < init_diff * CONVERGENCE_RATIO
                    ):
                    break

            last_slope = slope

        return xpp, ypp, slope, intercept, ppp

    def show_animation( self, save_file=None ):
        y = self.y

        for ax in self.axes:
            ax.plot( y )

        ax1, ax2 = self.axes
        xmin, xmax = ax2.get_xlim()
        ymin, ymax = ax2.get_ylim()
        w1 = 0.02
        w2 = 0.98
        xmin_ = xmin*(1-w1) + xmax*w1
        xmax_ = xmin*(1-w2) + xmax*w2
        ymin_ = ymin*(1-w1) + ymax*w1
        ymax_ = ymin*(1-w2) + ymax*w2

        for ax, lw in zip(self.axes, [5, 10]):
            ax.plot([xmin_, xmax_, xmax_, xmin_, xmin_], [ymin_, ymin_, ymax_, ymax_, ymin_], color='red', alpha=0.1, linewidth=lw)

        dual_amin_data = []
        for data1, data2 in zip(*self.anim_data_lists):
            dual_amin_data.append(data1+data2)

        # ani = animation.ArtistAnimation(self.fig, self.anim_data)
        ani = animation.ArtistAnimation(self.fig, dual_amin_data, interval=1000, blit=True, repeat_delay=1000 )

        if save_file is not None:
            print("saving anim to ", save_file)
            ani.save( save_file, writer="ffmpeg" )

        plt.show()



"""
    OptimizerUtils.py

    Copyright (c) 2021-2024, SAXS Team, KEK-PF
"""
MODEL_NAME_DICT = {
    "G0346" : "EGH",
    "G1100" : "SDM",
    "G2010" : "EDM",
}
def get_model_name(class_code):
    return MODEL_NAME_DICT[class_code]

METHOD_NAMES = ["BH", "NS", "MCMC", "SMC"]
def get_method_name():
    from molass_legacy._MOLASS.SerialSettings import get_setting
    return METHOD_NAMES[get_setting("optimization_method")]

IMPL_METHOD_NAMES = ["bh", "ultranest", "emcee", "pyabc", "pymc"]
def get_impl_method_name(nnn, method=None):
    if method is None:
        from molass_legacy._MOLASS.SerialSettings import get_setting
        method = get_setting("optimization_method")
    if method >= 2:
        r = method % 2
        method = (nnn + r) % 2
    return IMPL_METHOD_NAMES[method]

def show_peak_editor_impl(strategy_dialog, dialog, pe_proxy=None, pe_ready_cb=None, apply_cb=None, debug=True):
    from molass_legacy._MOLASS.SerialSettings import get_setting
    if debug:
        from importlib import reload
        import molass_legacy.SecSaxs.DataTreatment
        reload(molass_legacy.SecSaxs.DataTreatment)
        import molass_legacy.Peaks.PeakEditor
        reload(molass_legacy.Peaks.PeakEditor)
    from molass_legacy.SecSaxs.DataTreatment import DataTreatment
    from molass_legacy.Peaks.PeakEditor import PeakEditor

    parent = dialog.parent

    if pe_proxy is None:
        exact_num_peaks = strategy_dialog.get_num_peaks()
        strict_sec_penalty, correction, trimming, unified_baseline_type = strategy_dialog.get_options()

        treat = DataTreatment(route="v2", trimming=trimming, correction=correction, unified_baseline_type=unified_baseline_type)
        sd = dialog.serial_data
        pre_recog = dialog.pre_recog
        trimmed_sd = treat.get_trimmed_sd(sd, pre_recog)
        corrected_sd = treat.get_corrected_sd(sd, pre_recog, trimmed_sd)

        dialog.grab_set()   # temporary fix to the grab_release problem

        pe = PeakEditor(parent, trimmed_sd, treat.pre_recog, corrected_sd, treat, exact_num_peaks=exact_num_peaks, strict_sec_penalty=strict_sec_penalty)
        if pe_ready_cb is not None:
            pe_ready_cb(pe)

        pe.show()
        if not pe.applied:
            return

        settings = None
    else:
        pe_proxy.load_settings()
        trimmed_sd = pe_proxy.get_sd()
        treat = pe_proxy.get_treat()
        pe = pe_proxy

    from molass_legacy.Optimizer.InitialInfo import InitialInfo
    if debug:
        from importlib import reload
        import molass_legacy.Optimizer.FullOptDialog
        reload(molass_legacy.Optimizer.FullOptDialog)
    from molass_legacy.Optimizer.FullOptDialog import FullOptDialog

    optinit_info = InitialInfo(trimmed_sd, treat=treat, pe=pe)
    dialog.grab_set()   # temporary fix to the grab_release problem, to be inspected to eventually remove this
    dialog.fullopt_dialog = FullOptDialog(parent, dialog, optinit_info)
    if apply_cb is not None:
        apply_cb(dialog.fullopt_dialog)
    dialog.fullopt_dialog.show()

class OptimizerResult:
    def __init__(self, x=None, nit=None, nfev=None):
        self.x = x
        self.nit = nit
        self.nfev = nfev

if __name__ == '__main__':
    for nnn, method in [(0, 0), (0, 1), (5, 2), (5, 3)]:
        if method < 2:
            print((nnn, method), get_impl_method_name(nnn, method=method))
        else:
            for n in range(nnn):
                print((n, method), get_impl_method_name(n, method=method))



# coding: utf-8
"""
    CorrectedData.py

    Copyright (c) 2020-2021, SAXS Team, KEK-PF
"""
import logging
import numpy as np
from RawData import RawXray, RawUv
from Prob.ProbData import ProbData
from LPM import LPM_3d
from molass_legacy.Baseline.LambertBeer import BasePlane
from molass_legacy.SerialAnalyzer.ElutionCurve import ElutionCurve
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.Trimming.PreliminaryRg import get_default_angle_range_impl

N_SIGMA_TO_RISTRICT_RANGE = 4

class CorrectedXray(RawXray):
    def __init__(self, in_folder=None, xr=None, mf_adjust=True, average_width=None):
        self.logger = logging.getLogger(__name__)
        if xr is None:
            xr = RawXray(in_folder)
        self.initialize(xr, mf_adjust, average_width)

    def exclude_elution(self, points):
        """
        do not use this
        use RawXray.exclude_elution instead
        """
        assert False

    def initialize(self, xr, mf_adjust=True, average_width=None):
        self.xr = xr
        i = xr.get_row_index(0.02)
        eslice, xr_y = self.get_elution_range(xr, i, average_width)
        y = xr_y[eslice]
        ecurve = ElutionCurve(y, j0=xr.j0)
        aslice = self.get_angular_range(xr, eslice, ecurve)

        self.qvector = xr.qvector[aslice].copy()
        self.vector = self.qvector
        self.ecurve = ecurve
        restricted = xr.data[aslice,eslice].copy()
        lpm = LPM_3d(restricted)
        if mf_adjust:
            data = self.adjust_with_mf_baseplane(lpm.data, i, ecurve)
        else:
            data = lpm.data
        error = xr.error[aslice,eslice].copy()
        j0 = xr.j0
        self.set_data(slice(j0+eslice.start, j0+eslice.stop), data, error)
        self.excluded = xr.excluded

    def get_elution_range(self, xr, i, average_width=None):
        if average_width is None:
            average_width = get_setting('num_points_intensity')

        hw = average_width//2
        xr_y = np.average(xr.data[i-hw:i+hw+1,:], axis=0)
        prb = ProbData(xr_y)
        # prb.proof_plot()
        start, stop = prb.get_approx_peak_range(N_SIGMA_TO_RISTRICT_RANGE)
        return slice(start, stop), xr_y

    def get_angular_range(self, xr, eslice, ecurve):
        qlimit = xr.get_row_index()
        start, stop, pre_rg = get_default_angle_range_impl(xr.data[:,eslice], xr.error[:,eslice], ecurve, xr.qvector, qlimit, self.logger)
        return slice(start, stop)

    def adjust_with_mf_baseplane(self, data, index, ecurve):
        try:
            bp = BasePlane(data, index, ecurve, denoise=False)
            bp.solve()
            BP = bp.get_baseplane()
            data -= BP
        except:
            import logging
            from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
            logger = logging.getLogger(__name__)
            etb = ExceptionTracebacker()
            logger.warning("determining a BasePlane failed with %s", str(etb))
        return data

    def set_data(self, eslice, data, error):
        slice_ = get_setting('qmm_window_slice')
        if slice_ is None:
            data_ = data
            error_ = error
            j0 = eslice.start
        else:
            start = max(eslice.start, slice_.start)
            stop = min(eslice.stop, slice_.stop)
            start_ = start - eslice.start
            stop_ = stop - eslice.start
            data_ = data[:, start_:stop_]
            error_ = error[:, start_:stop_]
            j0 = start
            self.cut_ecurve(start_, stop_)

        self.data = data_
        self.error = error_
        self.j0 = j0

    def cut_ecurve(self, start, stop):
        pass

class CorrectedUv(RawUv):
    def __init__(self, in_folder=None, uv=None, mf_adjust=True):
        if uv is None:
            uv = RawUv(in_folder)
            assert uv.data is not None
        self.uv = uv

        i = uv.get_row_index(280)
        uv_y = uv.data[i,:]
        prb = ProbData(uv_y)
        # prb.proof_plot()
        start, stop = prb.get_approx_peak_range(N_SIGMA_TO_RISTRICT_RANGE)
        eslice = slice(start, stop)

        self.wvector = uv.wvector.copy()
        self.vector = self.wvector
        self.ecurve = None
        restricted = uv.data[:,eslice].copy()
        lpm = LPM_3d(restricted)
        if mf_adjust:
            data = self.adjust_with_mf_baseplane(lpm.data, i, uv_y[eslice])
        else:
            data = lpm.data
        j0 = uv.j0
        self.set_data(slice(j0+eslice.start, j0+eslice.stop), data)
        self.error = None

    def adjust_with_mf_baseplane(self, data, index, y):
        try:
            ecurve = ElutionCurve(y, j0=self.uv.j0)
            self.ecurve = ecurve
            bp = BasePlane(data, index, ecurve, denoise=False)
            bp.solve()
            BP = bp.get_baseplane()
            data -= BP
        except:
            import logging
            from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
            logger = logging.getLogger(__name__)
            etb = ExceptionTracebacker()
            logger.warning("determining a BasePlane failed with %s", str(etb))
        return data

    def set_data(self, eslice, data):
        slice_ = get_setting('qmm_window_slice_uv')
        if slice_ is None:
            data_ = data
            j0 = eslice.start
        else:
            start = max(eslice.start, slice_.start)
            stop = min(eslice.stop, slice_.stop)
            start_ = start - eslice.start
            stop_ = stop - eslice.start
            data_ = data[:, start_:stop_]
            j0 = start

        self.data = data_
        self.j0 = j0



import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation as animation

N = 200
n = N//2
r = np.arange(N)
rr = r**2
i = np.arange(N)
ii, jj = np.meshgrid(i, i)
labels = np.searchsorted(rr, (ii - n)**2 + (jj-n)**2)
print(labels.shape)
Lmax = np.max(labels)
print(labels[0:3,0:3])

canvas = np.zeros((N, N))

fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(14,7))

# ax2.imshow(labels)

x = np.linspace(0, np.pi*12, N + Lmax)

ims = []
for i in range(60):
    t = np.pi*i/3
    beam = 1 + np.sin(x[0:n] - t)
    canvas[100,0:n] = beam
    im = ax1.imshow(canvas, animated=True)
    ims.append([im])
    scattered = 1 + np.sin(x[n:] - t)
    im = ax2.imshow(scattered[labels], animated=True)
    ims.append([im])

ani = animation.ArtistAnimation(fig, ims, interval=50, blit=True)
fig.tight_layout()
plt.show()



# coding: utf-8
"""
    Pomegranate.Spike.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
from time import time
import asyncio
import numpy as np
from pomegranate import *
from molass_legacy.Peaks.ElutionModels import egh
import molass_legacy.KekLib.DebugPlot as plt
from .EghMixtureModel import EghMixtureModel

def normal_pdf(x, mu, sigma):
    return np.exp(-1/2*((x - mu)/sigma)**2)/(sigma*np.sqrt(np.pi))

def distributions_demo():
    x = np.arange(500)

    plt.push()
    fig, ax = plt.subplots()

    d = EghDistribution(200, 30, 5)
    ax.plot(x, d.probability(x))

    fig.tight_layout()
    plt.show()
    plt.pop()

def spike_demo():
    x = np.arange(500)
    y = np.zeros(len(x))
    for params in [
            # [0.6, 200, 40, 10],
            [0.6, 200, 30, 5],
            [1.0, 300, 30, 10],
            ]:
        y += egh(x, *params)

    spike_demo_impl([(x, y)], num_components=2)

def spike_demo_impl(xy_list, num_components=None, use_peaks=False):
    if use_peaks:
        from molass_legacy.Peaks.RobustPeaks import RobustPeaks

    models = []
    for x, y in xy_list:
        model = EghMixtureModel(x, y, num_components)
        models.append(model)

    if use_peaks:
        peaks_list = []
        for x, y in xy_list:
            rp = RobustPeaks(x, y)
            peaks_list.append(rp)

    for n in range(100):
        t0 = time()
        results = []
        for k, model in enumerate(models):
            x, y = xy_list[k]
            if n == 0:
                peaks = peaks_list[k] if use_peaks else None
                res = model.fit(peaks=peaks)
            else:
                res = model.fit()
            cy_list, ty = model.get_components(res, x, y)
            results.append((cy_list, ty))

        print("It took %.3g seconds." % (time()-t0))

        plt.push()
        fig, axes = plt.subplots(ncols=2, figsize=(14,7))

        for k, (x, y) in enumerate(xy_list):
            ax = axes[k]
            cy_list, ty = results[k]
            ax.plot(x, y)
            for cy in cy_list:
                ax.plot(x, cy, ':')
            ax.plot(x, ty, ':', color='red')

        fig.tight_layout()
        cont = plt.show()
        plt.pop()

        if not cont:
            break

def spike_demo_real(in_folder, num_components=3, correction=False, use_peaks=False):
    from molass_legacy.Batch.StandardProcedure import StandardProcedure
    sp = StandardProcedure()
    sp.load(in_folder, debug=False)
    if correction:
        sd = sp.get_corrected_sd()
    else:
        sd = sp.get_sd()
    uv_curve = sd.get_uv_curve()
    xr_curve = sd.get_xray_curve()
    xy_list = [(curve.x, curve.y) for curve in [uv_curve, xr_curve]]
    spike_demo_impl(xy_list, num_components=4, use_peaks=use_peaks)

async def get_content(n):
    await asyncio.sleep(3)
    return n + 1

async def f(n):
    content = await get_content(n)
    return content

def coroutine_spike():
    loop = asyncio.get_event_loop()
    v = loop.run_until_complete(f(1))
    print(v)



"""
    RateTheoryDemo.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np
from matplotlib.gridspec import GridSpec
import molass_legacy.KekLib.DebugPlot as plt
from DataUtils import get_in_folder
from molass_legacy.Models.ElutionCurveModels import EGHA, EMGA

def compare_models(in_folder, sd):
    from importlib import reload
    import Models.RateTheory.EDM
    reload(Models.RateTheory.EDM)
    from .RateTheory.EDM import guess, edm_impl

    ecurve = sd.get_xray_curve()
    x = ecurve.x
    y = ecurve.y

    model = EGHA()
    g_params = model.guess(y, x=x)
    ret = model.fit(y, g_params, x=x)
    y_egh = model.eval(ret.params, x=x)

    model = EMGA()
    g_params = model.guess(y, x=x)
    ret = model.fit(y, g_params, x=x)
    y_emg = model.eval(ret.params, x=x)

    params = guess(x, y)
    print("params=", params)

    with plt.Dp():
        fig = plt.figure(figsize=(18,7))
        fig.suptitle("Model Comparison using %s" % get_in_folder(in_folder), fontsize=20)
        gs = GridSpec(4,3)
        ax11, ax12, ax13, ax21, ax22, ax23 = [ fig.add_subplot(gs_) for gs_ in (gs[0:3,0], gs[0:3,1], gs[0:3,2], gs[3,0], gs[3,1], gs[3,2])]
        ax11.set_title("EGH", fontsize=16)
        ax12.set_title("EMG", fontsize=16)
        ax13.set_title("EDM", fontsize=16)

        def draw_rmsd(ax1, rmsd):
            tx = np.mean(ax1.get_xlim())
            ty = np.mean(ax1.get_ylim())
            text = "rmsd=%.2g" % rmsd
            ax1.text(tx, ty, text, va="center", ha="center", fontsize=40, alpha=0.3)

        rmsds = []

        ax11.plot(x, y)
        ax11.plot(x, y_egh)
        div = y_egh - y
        ax21.plot(x, div)
        rmsd = np.sqrt(np.mean(div**2))
        rmsds.append(rmsd)
        draw_rmsd(ax11, rmsd)

        ax12.plot(x, y)
        ax12.plot(x, y_emg)
        div = y_emg - y
        ax22.plot(x, div)
        rmsd = np.sqrt(np.mean(div**2))
        rmsds.append(rmsd)
        draw_rmsd(ax12, rmsd)

        ax13.plot(x, y)
        y_ = edm_impl(x, *params)
        ax13.plot(x, y_)
        div = y_ - y
        ax23.plot(x, div)
        rmsd = np.sqrt(np.mean(div**2))
        rmsds.append(rmsd)
        draw_rmsd(ax13, rmsd)

        ylims = []
        for ax1 in ax21, ax22, ax23:
            ylims.append(ax1.get_ylim())

        ylims = np.array(ylims)
        ymin = np.min(ylims[:,0])
        ymax = np.max(ylims[:,1])

        for ax1 in ax21, ax22, ax23:
            ax1.set_ylim(ymin, ymax)

        m = np.argmin(rmsds)
        for k, axes in enumerate(((ax11, ax21), (ax12, ax22), (ax13, ax23))):
            if k == m:
                for ax1 in axes:
                    ax1.set_facecolor("peachpuff")

        fig.tight_layout()
        plt.show()

def edm_inspect(in_folder, sd):
    import warnings
    warnings.filterwarnings("ignore")
    from matplotlib.widgets import Slider, Button
    from importlib import reload
    import Models.RateTheory.EDM
    reload(Models.RateTheory.EDM)
    from .RateTheory.EDM import guess, edm_impl

    ecurve = sd.get_xray_curve()
    x = ecurve.x.copy()
    y = ecurve.y

    x += 0

    params = guess(x, y)
    print("params=", params)
    fullparams = np.concatenate([params, [0, 0.0001, 2.0, 30.0, 30.0]])

    slider_specs = [    ("t0", -50, 100, fullparams[0]),
                        ("u", 0, 2, fullparams[1]),
                        ("a", 0, 2, fullparams[2]),
                        ("b", -4, 1, fullparams[3]),
                        ("e", 0, 2, fullparams[4]),
                        ("Dz", 0, 1, fullparams[5]),
                        ("cinj", 0, 10, fullparams[6]),
                        ("cinit", 0, 1, fullparams[7]),
                        ("c0", 0, 3, fullparams[8]),
                        ("tinj", 0, 3, fullparams[9]),
                        ("L", 0, 35, fullparams[10]),
                        ("z", 0, 35, fullparams[11]),
                        ]

    in_folder = get_in_folder(in_folder)
    with plt.Dp():
        fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(24,10))
        dp = plt.get_dp()

        fig.suptitle("EDM Params Inspection for %s" % in_folder, fontsize=32)
        ax2.set_axis_off()

        ax1.plot(x, y)
        edm_line, = ax1.plot(x, edm_impl(x, *params))

        def slider_update(k, val):
            print([k], "slider_update", val)
            params[k] = val
            y_ = edm_impl(x, *params)
            edm_line.set_data(x, y_)
            dp.draw()

        slider_axes = []
        sliders = []
        for k, (label, valmin, valmax, valinit) in enumerate(slider_specs, start=0):
            ax_ = fig.add_axes([0.6, 0.8 - 0.06*k, 0.3, 0.03])
            slider  = Slider(ax_, label=label, valmin=valmin, valmax=valmax, valinit=valinit)
            slider.label.set_size(16)
            slider.on_changed(lambda val, k_=k: slider_update(k_, val))
            slider_axes.append(ax_)
            sliders.append(slider)

        def reset_slider_valules(event):
            for slider in sliders:
                slider.reset()
            dp.draw()

        axreset = fig.add_axes([0.9, 0.0, 0.1, 0.075])
        breset = Button(axreset, 'Reset')
        breset.label.set_fontsize(16)
        breset.on_clicked(reset_slider_valules)

        ax1.tick_params(axis='x', labelsize=16)
        ax1.tick_params(axis='y', labelsize=16)

        fig.tight_layout()
        plt.show()




# coding: utf-8
"""
    MXD.Spike.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
from time import time
import numpy as np
from sklearn.mixture import GaussianMixture
from sklearn.cluster import KMeans
from molass_legacy.Peaks.ElutionModels import egh
from .FastDecomposer import FastDecomposer
import molass_legacy.KekLib.DebugPlot as plt

def spike_demo():
    x = np.arange(500)
    y = np.zeros(len(x))
    for params in [
            # [0.6, 200, 40, 10],
            [0.6, 200, 30, 5],
            [1.0, 300, 30, 10],
            ]:
        y += egh(x, *params)

    spike_demo_impl(x, y, num_compoments=2)

def spike_demo_impl(x, y, num_compoments=None):
    from Prob.GaussianMixture import hist_to_source
    from Prob.EghMixture import EghMixture
    from Prob.EghMixtureUtils import get_components

    t0 = time()
    sy = hist_to_source(x, y)
    mm = EghMixture(num_compoments)
    mm.fit(np.expand_dims(sy,1))
    print("it took %.3g seconds." % (time() - t0))
    cy_list, ty = get_components(x, y, mm)

    try:
        t0 = time()
        fd = FastDecomposer(num_compoments)
        fd.fit(x, y)
        print("it took %.3g seconds." % (time() - t0))
        cy_list2, ty2 = fd.get_components()
    except:
        print("FastDecomposer failed")
        mm = EghMixture(num_compoments)
        mm.fit(np.expand_dims(sy,1))
        cy_list2, ty2 = get_components(x, y, mm)

    plt.push()
    fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(14,7))
    ax1.plot(x, y)
    ax1.plot(x, ty, ':')
    for cy in cy_list:
        ax1.plot(x, cy, ':')

    ax2.plot(x, y)
    ax2.plot(x, ty2, ':')
    for cy in cy_list2:
        ax2.plot(x, cy, ':')

    fig.tight_layout()
    plt.show()
    plt.pop()

def spike_demo_real(in_folder):
    from molass_legacy.Batch.StandardProcedure import StandardProcedure
    sp = StandardProcedure()
    sp.load(in_folder, debug=False)
    sd = sp.get_sd()
    xr_curve = sd.get_xray_curve()
    spike_demo_impl(xr_curve.x, xr_curve.y, num_compoments=5)



"""
    LrfInfo.py

    Copyright (c) 2020-2023, SAXS Team, KEK-PF
"""
import numpy as np
from bisect import bisect_right
from SimpleGuinier import SimpleGuinier

DEBUG = False
if DEBUG:
    import logging

class LrfInfo:
    def __init__(self, opt_info, A, lrfE):
        if DEBUG:
            self.logger = logging.getLogger(__name__)
        self.num_iterations = opt_info[0]
        nth, cdl_list = opt_info[3]
        self.need_bq_ = cdl_list[nth] > 1
        qv = opt_info[4]
        self.boundary = opt_info[5]
        self.boundary_j = None if self.boundary is None else bisect_right(qv, self.boundary)
        self.data = opt_info[6]

        A_data = np.array( [qv, A, lrfE[0]] ).T

        sg = SimpleGuinier(A_data)
        self.Rg = Rg = sg.Rg
        self.sg = sg
        self.basic_quality = sg.basic_quality
        if DEBUG:
            Rg_ = str(Rg) if Rg is None else "%.3g" % Rg
            self.logger.info("Rg=%s with basic_quality=%.3g", Rg_, self.basic_quality)

    def need_bq(self):
        return self.need_bq_



# coding: utf-8
"""
    Peaks.WindowMapping.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
import numpy as np
USE_BASINHOPPING = False
if USE_BASINHOPPING:
    from scipy.optimize import basinhopping
else:
    from scipy.optimize import minimize
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.QuickAnalysis.PeakUtils import recognize_peaks
from molass_legacy.Peaks.ElutionModels import egh

def demo(parent, in_folder):
    from molass_legacy._MOLASS.SerialSettings import set_setting
    from molass_legacy.Batch.StandardProcedure import StandardProcedure
    from molass_legacy.Elution.CurveUtils import simple_plot

    sp = StandardProcedure()
    sp.load(in_folder)
    sd = sp.get_sd(whole=True)

    uv_curve = sd.get_uv_curve()
    xr_curve = sd.get_xray_curve()


    uv_x = uv_curve.x
    uv_y = uv_curve.y
    uv_spline = uv_curve.spline

    x = xr_curve.x
    y = xr_curve.y
    xr_peaks = recognize_peaks(x, y)

    model_y = np.zeros(len(y))
    for h, mu, sigma, tau in xr_peaks:
        cy = egh(x,  h, mu, sigma, tau)
        model_y += cy

    fx = x[0]
    tx = x[-1]

    def debug_plot(p, add_info=None):
        from DataUtils import get_in_folder

        a, b, fx, tx, scale = p

        plt.push()
        fig, (ax1, ax2, ax3) = plt.subplots(ncols=3, figsize=(18,6))
        fig.suptitle("Debug Plot for %s" % get_in_folder(in_folder=in_folder), fontsize=20)
        ax1.set_title("UV Elution", fontsize=16)
        ax2.set_title("Xray Elution", fontsize=16)

        simple_plot(ax1, uv_curve)
        simple_plot(ax2, xr_curve, color="orange")

        ty = np.zeros(len(y))
        for h, mu, sigma, tau in xr_peaks:
            cy = egh(x,  h, mu, sigma, tau)
            ty += cy
            ax2.plot(x, cy, ':')

        ax2.plot(x, ty, ':')

        ymin1, ymax1 = ax1.get_ylim()
        ax1.set_ylim(ymin1, ymax1)

        ymin2, ymax2 = ax2.get_ylim()
        ax2.set_ylim(ymin2, ymax2)
        for x_ in [fx, tx]:
            uv_x_ = a*x_ + b
            ax1.plot([uv_x_, uv_x_], [ymin1, ymax1], ':', color="gray")
            ax2.plot([x_, x_], [ymin2, ymax2], ':', color="gray")

        if add_info is not None:
            a, b, jfx, jtx, ifx, itx, residual, range_score, order_penalty, deviation_penalty, xr_range_penalty, uv_range_penalty, wx, wy, wbase = add_info

            print("a, b=", a, b)
            print("(jfx, jtx), (ifx, itx)=", (jfx, jtx), (ifx, itx))
            ax1.plot(wx, wy)
            ax1.plot(wx, wbase, color="red")

            score_list = [residual, range_score, order_penalty, deviation_penalty, xr_range_penalty, uv_range_penalty]
            score_names = "residual, range_score, order_penalty, deviation_penalty, xr_range_penalty, uv_range_penalty"

            fv = np.sum(score_list)
            ax3.set_title("fv=%.3g" % fv, fontsize=16)

            ax3.yaxis.tick_right()
            ax3.yaxis.set_label_position("right")
            y_pos = np.arange(len(score_list))
            ax3.barh(y_pos, score_list)
            ax3.set_yticks(y_pos)
            ax3.invert_yaxis() 
            ax3.set_yticklabels(score_names.split(", "))

        fig.tight_layout()
        ret = plt.show()
        plt.pop()
        return ret

    a_init = len(uv_x)/len(x)
    b_init = 0
    scale_init = np.max(uv_y)/np.max(y)
    wh = np.where(model_y > xr_curve.max_y*0.01)[0]
    init_params = (a_init, b_init, x[wh[0]], x[wh[-1]], scale_init)

    VERY_LARGE_VALUE = 1e6
    min_a = a_init*0.7
    max_a = a_init*1.4
    max_x = x[-1]
    min_width = max_x*0.2
    min_s = scale_init*0.7
    max_s = scale_init*1.4

    debug = True

    def objective(p, plot=False):
        nonlocal debug

        a, b, fx, tx, scale = p

        if fx < 0 or tx > max_x or tx - fx < min_width or a < min_a or a > max_a or scale < min_s or scale > max_s:
            jfx, jtx, ifx, itx = 0, 0, 0, 0
            residual = VERY_LARGE_VALUE
            range_score = VERY_LARGE_VALUE
            order_penalty = VERY_LARGE_VALUE
            deviation_penalty = VERY_LARGE_VALUE
            xr_range_penalty = VERY_LARGE_VALUE
            uv_range_penalty = VERY_LARGE_VALUE
            wx = wy = wbase = np.zeros(len(x))
        else:
            ifx = max(0, int(np.round(fx)))
            itx = min(len(x)-1, int(np.round(tx)))
            wx = a*x[ifx:itx] + b
            wy = uv_spline(wx)
            n = len(wy)
            wbase = wy[0] + np.arange(n)*1/(n-1)*(wy[-1] - wy[0])
            wy_ = wy - wbase - scale*y[ifx:itx]

            residual = np.sum(wy_**2)
            # range_score = 1e1*(y[ifx]+y[itx])**2/(tx - fx)**2
            jfx = max(0, int(np.round(a*fx + b)))
            jtx = min(len(uv_x) - 1, int(np.round(a*tx + b)))

            range_score = 1/np.sum(model_y[ifx:itx])
            order_penalty = VERY_LARGE_VALUE*(min(0, tx - fx)**2 + min(0, a)**2)
            deviation_penalty = 0
            xr_range_penalty = VERY_LARGE_VALUE*(min(0, fx - x[0])**2 + max(0, tx - x[-1])**2)
            uv_range_penalty = VERY_LARGE_VALUE*(min(0, wx[0] - uv_x[0])**2 + max(0, wx[-1] - uv_x[-1])**2)

        fv = residual + range_score + order_penalty + deviation_penalty + xr_range_penalty + uv_range_penalty

        if debug or plot:
            print("residual=", residual)
            print("range_score=", range_score)
            print("xr_range_penalty=", xr_range_penalty)
            print("uv_range_penalty=", uv_range_penalty)
            debug = debug_plot(p, add_info=(a, b, jfx, jtx, ifx, itx, residual, range_score, order_penalty, deviation_penalty, xr_range_penalty, uv_range_penalty, wx, wy, wbase))

        return fv

    if USE_BASINHOPPING:
        minimizer_kwargs = {"method": "Nelder-Mead"}
        ret = basinhopping(objective, init_params, niter=20, minimizer_kwargs=minimizer_kwargs)
    else:
        ret = minimize(objective, init_params, method="Nelder-Mead")

    objective(ret.x, plot=True)



"""
    SecTheory.MwRgInfo.py

    Copyright (c) 2022-2023, SAXS Team, KEK-PF
"""
import os
import numpy as np
from scipy.stats import linregress
import io
import re

TABLE1 = """
Table 1
Light scattering results for standard globular proteins.

Protein             Oligomeric state  Mw(kDa)   Rh(nm)  Rg(nm)
------------------------------------------------------------
Aprotinin               Monomer         6.8     1.35    1.04
Cytochrome C            Monomer         12      1.77    1.37
α-Lactalbumin          Monomer         14.3    1.91    1.49
Myoglobin               Monomer         14.2    2.12    1.64
Carbonic anhydrase      Monomer         29.2    2.35    1.82
Trypsin inhibitor       Monomer         20.5    2.47    1.91
β-Lactoglobulin        Monomer         20.1    2.64    2.04
Ovalbumin               Monomer         42.5    2.98    2.31
Bovine serum albumine   Monomer         67.1    3.56    2.76
Enolase (yeast)         Dimer           79.5    3.57    2.76
Enolase (rabbit)        Dimer           86.4    3.65    2.83
Transferrin             Monomer         76.9    4.02    3.11
Alcohol dehydrogenase   Tetramer        144     4.5     3.48
BSA dimer               Dimer           137.1   3.68    3.62
Aldolase (rabbit)       Tetramer        155     4.77    3.69
------------------------------------------------------------

from
https://www.ncbi.nlm.nih.gov/pmc/articles/PMC4603275/pdf/j-48-01604.pdf

Molecular weight - gyration radius relation of globular proteins:
a comparison of light scattering, small-angle X-ray scattering
and structure-based data

by Detlef-M. Smilgies and Ewa Folta-Stogniew


Table 1
Experimental and computed structural parameters of the studied proteins.

Protein                PDB used       MM(kDa)            Rg(nm)
-------------------------------------------------------------
Ribonuclease A          1FS3            13.7      _      1.58
Lysozyme                1LYZ            14.3      _      1.43
Chymotrypsinogen A      2CGA            25.0      _      1.85
Carbonic anhydrase      1V9E            29.0      _      2.08
Ovalbumin               1OVA            45.0      _      2.66
BSA                     1N5U            66.0      _      2.99
Alcohol dehydrogenase   1JVB           150        _      3.27
Aldolase                1ZAH           158        _      3.51
Glucose isomerase       1OAD           173        _      3.25
β-Amylase              1FA2           200        _      4.22
Catalase                4BLC           232        _      3.84
Apoferritin             1IER           440        _      7.05
Thyroglobulin           _              669        _      7.56
-------------------------------------------------------------

from
https://journals.iucr.org/j/issues/2007/s1/00/sm6016/sm6016.pdf
Accuracy of molecular mass determination of
proteins in solution by small-angle X-ray scattering

by Efstratios Mylonasa and Dmitri I. Svergun


QUERY: Molecular Weight per Deposited Model >= 800, etc.

Protein                PDB used       MM(kDa)            Rg(nm)
-------------------------------------------------------------
GROEL                  1AON           841.75      _     6.972
RIBOSOME               1FFK          1321.3       _     6.928
OXIDOREDUCTASE         1GQ2          1015.3       _     8.430
LIGASE                 1HTO          1296.2       _     8.260
-------------------------------------------------------------

ENDOCYTOSIS/EXOCYTOSIS 1XI4          1683.0       _    14.80
RIBOSOME               2J28          1411.1       _     7.647

by PDB and CRYSOL3.0
"""

MWRG_CSV_PATH = os.path.join(os.path.dirname(__file__), "mwrg.csv")

class MwRgInfo:
    def __init__(self):
        table = io.StringIO(TABLE1)
        is_data_line = 0
        names = []
        data = []
        for line in table:
            if line.find("----") >= 0:
                is_data_line = 1 - is_data_line
                continue
            if is_data_line:
                words = re.split(r"\s+", line[:-1])
                mw = float(words[-3])
                rg = float(words[-1])*10
                data.append((mw, rg))
                names.append(words[-4])

        data = np.array(data)

        if os.path.exists(MWRG_CSV_PATH):
            extra_data = []
            with open(MWRG_CSV_PATH) as fh:
                for line in fh:
                    words = line[:-1].split(',')
                    values = [float(s) for s in words[1:]]
                    if values[1] > 0:   # Rg > 0
                        extra_data.append(values)
                        names.append(words[0])
            extra_data = np.array(extra_data)
        else:
            extra_data = np.array([]).reshape((0,2))
        self.mws = np.concatenate([data[:,0], extra_data[:,0]/1000])
        self.rgs = np.concatenate([data[:,1], extra_data[:,1]])
        self.names = names

        self.result = linregress(np.log(self.mws), np.log(self.rgs))

    def compute_rg(self, mw, n_sigma=0):
        slope, intercept, rvalue, pvalue, stderr = self.result
        intercept_stder = self.result.intercept_stderr

        a = intercept + n_sigma*intercept_stder
        b = slope + n_sigma*stderr
        return np.exp(a + b*np.log(mw))

    def compute_mw(self, rg):
        slope, intercept = self.result[0:2]
        return np.exp((np.log(rg) - intercept)/slope)

mwrg_info = None

def get_mwrg_info(devel=False):
    global mwrg_info
    if mwrg_info is None or devel:
        mwrg_info = MwRgInfo()
    return mwrg_info



"""
    SaxsCurveUtils.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np

def percentile_normalize(y, percentile=98):
    # select 98-percentille point as peak-point to avoid abnormal max
    # see also SimpleGuinier.py
    head_len = len(y)//4
    kth = int(head_len*percentile/100)
    arg_hy98 = np.argpartition(y[0:head_len], kth)
    m = arg_hy98[kth]
    return y/y[m]



"""
    UV.PlainCurveUtils.py

    Copyright (c) 2022-2023, SAXS Team, KEK-PF
"""
from bisect import bisect_right
from molass_legacy._MOLASS.SerialSettings import get_setting

def get_flat_wavelength_simple():
    wavelength = get_setting('zero_absorbance')

    zero_absorbance_auto = get_setting('zero_absorbance_auto')
    if zero_absorbance_auto:
        flat_wavelength = get_setting('flat_wavelength')
        if flat_wavelength is not None:
            wavelength = flat_wavelength

    return wavelength

def get_flat_info_impl(wvector):
    w = get_flat_wavelength_simple()
    i = bisect_right(wvector, w)
    max_i = len(wvector) - 1
    i = min(i, max_i)
    if i == max_i:
        w = wvector[i]
    return w, i

def get_flat_info(sd):
    w, i = get_flat_info_impl(sd.lvector)
    return w, i, sd.conc_array[i,:]

def get_flat_wavelength(wvector=None):
    if wvector is None:
        w = get_flat_wavelength_simple()
    else:
        w, i = get_flat_info_impl(wvector)
    return w

def get_both_wavelengths(wvector=None):
    w = get_setting('absorbance_picking')
    return w, get_flat_wavelength(wvector)



# coding: utf-8
"""
    Debug.py

    Copyright (c) 2020-2022, SAXS Team, KEK-PF
"""

import os

def debug_log_path():
    this_dir = os.path.dirname(os.path.abspath(__file__))
    return os.path.abspath(this_dir + '/../../log/debug.log')

class DebugInfo:
    def __init__(self, debug=0):
        path = debug_log_path()
        self.debug = debug
        if self.debug:
            import logging
            self.logger = logging.getLogger()
            self.formatter = logging.Formatter('%(asctime)s %(message)s', '%Y-%m-%d %H:%M:%S')
            self.fileh = logging.FileHandler(path, 'w')
            self.fileh.setFormatter(self.formatter)
            self.logger.addHandler(self.fileh)
        else:
            self.logger = None

MOLASS_DEBUG_ARGS = "MOLASS_DEBUG_ARGS"

def set_debug_args(args):
    os.environ[MOLASS_DEBUG_ARGS] = args

class DebugArgs:
    def __init__(self):
        args = os.environ.get(MOLASS_DEBUG_ARGS)
        if args is not None:
            self.parse(args)
        self.args = args

    def is_active(self):
        return self.args is not None

    def parse(self, args):
        import re
        arg_re = re.compile(r"\s*(\w+)\s*=\s*(\S+)\s*")
        for term in args.split(','):
            m = arg_re.match(term)
            if m:
                name = m.group(1)
                value = eval(m.group(2))
                # print(name, value)
                self.__setattr__(name, value)



"""
    EmbedDialog.py

    edit this code while keeping the caller code unchanged

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import os
import sys

def open_dialog(callback, *args, **kwargs):
    from molass_legacy.KekLib.OurTkinter import Tk, Dialog

    begin_main_loop = False
    parent = kwargs.pop('parent', None)
    if parent is None:
        parent = Tk.Tk()
        begin_main_loop = True

    class EmbedDialog(Dialog):
        def __init__(self, parent, callback, *args, **wkargs):
            self.callback = callback
            self.args = args
            self.kwargs = wkargs
            Dialog.__init__(self, parent, "EmbedDialog", visible=False)

        def show( self ):
            self._show()

        def body(self, body_frame):
            button = Tk.Button(body_frame, text="Run", command=self.run)
            button.pack(padx=50, pady=30)

        def run(self):
            self.callback(*self.args, **self.kwargs)

    dialog = EmbedDialog(parent, callback, *args, **kwargs)

    def show_dialog():
        dialog.show()
        if begin_main_loop:
            parent.quit()

    if begin_main_loop:
        parent.withdraw()
        parent.after(0, show_dialog)
        parent.mainloop()
    else:
        show_dialog()

if __name__ == '__main__':
    this_dir = os.path.dirname(os.path.abspath(__file__))
    lib_dir = os.path.dirname(this_dir)
    sys.path.append(lib_dir)
    from DevelUtils.LegacyImporters import import_legacy_modules
    import_legacy_modules()

    def test_cb(*args, **kwargs):
        print(*args)
        print(**kwargs)

    open_dialog(test_cb, 1, 2, 3)



# coding: utf-8
"""

    ExcelLogChart.py

    Copyright (c) 2017, Masatsuyo Takahashi, KEK-PF

"""
import numpy    as np
from ExcelCOM   import xlValue, xlCategory, xlNone, CellAddr, msoFalse
from Unicode    import to_superscripts

msoChartFieldRange  = 7
xlLabelPositionLeft = -4131

class ExcelLogChart:
    def __init__( self, chart, worksheet ):
        self.chart  = chart
        self.x_axis = chart.Axes( xlCategory )
        self.y_axis = chart.Axes( xlValue )
        self.ws     = worksheet

    def super_ticklabel( self, tick ):
        return '10' + to_superscripts( '%d' % tick )

    def plain_ticklabel( self, tick ):
        return '10^%d' % tick

    def change_y_labels( self, superscripts=False ):
        # self.x_axis.TickLabelPosition = xlNone
        self.y_axis.TickLabelPosition = xlNone

        print( 'MajorUnit=', self.y_axis.MajorUnit )
        print( 'MinimumScale=', self.y_axis.MinimumScale )
        print( 'MaximumScale=', self.y_axis.MaximumScale )

        minscale = int( np.floor( np.log10( self.y_axis.MinimumScale) ) )
        maxscale = int( np.ceil ( np.log10( self.y_axis.MaximumScale) ) )

        y_ticks  = np.arange( minscale, maxscale+1 )
        label_x = [ 0.0 ] * len(y_ticks)
        label_y = [ np.power( 10.0, float(tick) ) for tick in y_ticks ]
        if superscripts:
            ticklabel = self.super_ticklabel
        else:
            ticklabel = self.plain_ticklabel
        label_t = [ ticklabel(tick) for tick in y_ticks ]
        # print( 'label_x=', label_x )
        # print( 'label_y=', label_y )
        # print( 'label_t=', label_t )

        usedrange   = self.ws.UsedRange
        work_row    = usedrange.Row + usedrange.Rows.Count + 5
        work_col    = usedrange.Column + usedrange.Columns.Count
        work_col1   = work_col+1
        work_col2   = work_col+2
        work_col3   = work_col+3

        print( 'work_row=', work_row )

        for i in range( len(y_ticks) ):
            self.ws.Cells( work_row + i, work_col1 ).Value = label_x[i]
            self.ws.Cells( work_row + i, work_col2 ).Value = label_y[i]
            self.ws.Cells( work_row + i, work_col3 ).Value = label_t[i]

        series = self.chart.SeriesCollection().NewSeries()
        minrow = work_row
        maxrow = work_row + len(y_ticks)
        series.Name = ""
        series.XValues  = self.reference( (minrow,work_col1), (maxrow,work_col1) )
        series.Values   = self.reference( (minrow,work_col2), (maxrow,work_col2) )
        series.Format.Line.Visible = msoFalse
        series.ApplyDataLabels()
        datalabels = series.DataLabels()
        datalabels.ShowRange = True
        datalabels.ShowValue = False
        labeldata_ref = self.reference( (minrow,work_col3), (maxrow,work_col3) )
        datalabels.Format.TextFrame2.TextRange.InsertChartField( msoChartFieldRange, labeldata_ref, 0 )
        datalabels.Position = xlLabelPositionLeft

    def get_plotarea( self ):
        plotarea = self.chart.PlotArea
        return ( plotarea.Left, plotarea.Top, plotarea.Width, plotarea.Height )

    def change_plotarea( self, left=None, top=None, width=None, height=None ):
        plotarea = self.chart.PlotArea
        if left is not None:
            plotarea.Left = left
        if top is not None:
            plotarea.Top = top
        if width is not None:
            plotarea.Width = width
        if height is not None:
            plotarea.Height = height

    def get_yaxis_title_area( self ):
        yaxis_title = self.chart.Axes(xlValue).AxisTitle
        return ( yaxis_title.Left, yaxis_title.Top, yaxis_title.Width, yaxis_title.Hieght )

    def change_yaxis_title_area( self, left=None, top=None, width=None, height=None ):
        yaxis_title = self.chart.Axes(xlValue).AxisTitle
        if left is not None:
            yaxis_title.Left = left
        if top is not None:
            yaxis_title.Top = top
        if width is not None:
            yaxis_title.Width = width
        if height is not None:
            yaxis_title.Height = height

    def reference( self, cell_f, cell_t ):
        ref = "='%s'!%s:%s" % ( self.ws.Name, CellAddr(cell_f), CellAddr(cell_t) )
        # print( 'ref=', ref )
        return ref



# coding: utf-8
"""
    NaturalSort.py

    Copyright (c) 2020, Masatsuyo Takahashi, KEK-PF
"""
import re

"""
    ナチュラルソート
    https://qiita.com/peg/items/8bf13cff8259ab464d8b
"""
def natural_sorted(l):
    def alphanum_key(s):
        return [int(c) if c.isdecimal() else c for c in re.split('([0-9]+)', s)]
    return sorted(l, key=alphanum_key)

def natural_sorted_dict_items(d):
    def alphanum_key(pair):
        ret = []
        for p in pair:
            ret += [int(c) if c.isdecimal() else c for c in re.split('([0-9]+)', p)]
        return ret
    return sorted(d.items(), key=alphanum_key)



"""
    QuickAnalysis.Homopeak.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np

class Homopeakness:
    def __init__(self, predec):
        self.sd = predec.sd
        self.sd_copy = predec.sd_copy
        self.xr_curve = predec.xr_curve
        self.xr_peaks = predec.xr_peaks
        self.xr_proportions = predec.get_xr_proportions()
        self.recognize_peak_groups()

    def recognize_peak_groups(self):
        groups = []
        prop_indeces = []
        start = 0
        stop_points = self.xr_curve.boundaries + [len(self.xr_curve.x)]
        for i, sp in enumerate(stop_points):
            print([i], sp)
            members = []
            indeces = []
            for j, peak in enumerate(self.xr_peaks[start:]):
                # peal[1] : mu
                if peak[1] < sp:
                    members.append(peak)
                    indeces.append(start + j)
                else:
                    break
            groups.append(np.array(members))
            prop_indeces.append(indeces)
            start += len(members)
        self.groups = groups
        self.prop_indeces = prop_indeces

    def get_homopeakness_scores(self):
        from SimpleGuinier import SimpleGuinier
        D, E, qv, _ = self.sd_copy.get_xr_data_separate_ly()
        x0 = self.sd_copy.xr_j0
        print("sd.shape=", self.sd.conc_array.shape, "sd_copy.shape=", self.sd_copy.conc_array.shape,  "x0=", x0)
        for k, (pinfo, members, indeces) in enumerate(zip(self.xr_curve.peak_info, self.groups, self.prop_indeces)):
            props = self.xr_proportions[indeces]
            m = np.argmax(props)
            h, mu, sigma, tau = self.xr_peaks[indeces[m]]
            rgs = []
            mu_ = mu - x0
            for px in mu_, mu_-sigma, mu_+sigma:
                j = int(round(px))
                slice_ = slice(j-2, j+3)
                y = np.sum(D[:,slice_], axis=1)
                ye = np.sum(E[:,slice_], axis=1)
                sg = SimpleGuinier(np.array([qv,y,ye]).T)
                rgs.append(sg.Rg)
            print([k], (pinfo, members[:,1], indeces, props, m), rgs)



"""
module for operations on moments
"""

import numpy as np
import cvxpy
from scipy.linalg import hankel
from .discrete_rv import DiscreteRV
import warnings

def empirical_moment(samples, degree):
    """ Compute empirical moments of samples
    Args:
    samples: x=(x1...xn)
    degree: L

    Returns
    matrix M of size L*n
    each row is the moments of (x1...xn) (start from the first degree to degree L)
    """
    m_raw = np.empty((degree, len(samples)))
    m_raw[0, :] = samples
    for i in range(1, degree):
        m_raw[i, :] = m_raw[i-1, :] * samples

    return m_raw

def hermite_transform_matrix(degree, sigma=1):
    """ Hermite transformation
    Let x=(x,...,x^k)', then Ax+b=(g_1(x,sigma),...,g_k(x,sigma))
    g_k(x,sigma)=sigma^k*H_k(x/sigma), g_k(x,1)=H_k(x), the usual Hermite polynomial

    Args:
    degree: int
    highest degree k

    Return:
    tuple (A,b): A is a matrix of shape (k,k), b is a vector of shape (k,1)
    """
    length = degree+1
    var = sigma*sigma
    mat = np.zeros((length, length))
    if length > 0:
        prepre = np.zeros(length)
        prepre[0] = 1
        mat[0, :] = prepre
    if length > 1:
        pre = np.zeros(length)
        pre[1] = 1
        mat[1, :] = pre
    for k in range(2, length):
        # recursion: H_{n+1}(x) = x * H_n(x) - n * H_{n-1}(x)
        # => g_{n+1}(x,s) = x * g_n(x,s) - n * s^2 * g_{n-1}(x,s)
        coeffs = np.roll(pre, 1) - prepre*(k-1)*var
        mat[k, :] = coeffs
        prepre = pre
        pre = coeffs

    return (mat[1:, 1:], mat[1:, 0].reshape((degree, 1)))


def transform(mat, x_var):
    """
    Compute a linear tranformation Ax+b

    mat: Tuple (A,b)
    A: matrix of shape (k,k)
    b: matrix of shape (k,1)

    x_var: variable x
    array of length k

    Returns:
    array of length k, y=Ax+b
    """
    k = len(mat)
    x_var = x_var.reshape((k, 1))
    y_var = np.dot(mat[0], x_var)+mat[1]
    return y_var.reshape(k)



def projection(moments, interval=None, wmat=None):
    """ project to a valid moment sequence on interval [a,b]

    Args:
    moments: a sequence of estimated moments, starting from degree 1
    interval [a,b]: range of distributions, default [-1,1]
    wmat: weighting matrix, default identity matrix

    Returns:
    a sequence of valid moments on [a,b], starting from degree 1 (default [a,b]=[-1,1])
    minimize (moments-x)' * wmat * (moments-x), subject to x is a valid moment sequence
    """
    if interval is None:
        interval = [-1, 1]

    length = len(moments)
    if length == 0:
        return moments
    if length == 1:
        moments[0] = max(interval[0], min(interval[1], moments[0]))
        return moments


    # preliminary filtering of moments based on range
    r_max = max(abs(interval[0]),abs(interval[1]))
    m_max = 1
    for i in range(len(moments)):
        m_max *= r_max
        if moments[i] > m_max:
            moments[i] = m_max
        elif moments[i] < -m_max:
            moments[i] = -m_max    

    # SDP for further projection
    variables = cvxpy.Variable(length) # variables [m_1,m_2,...,m_n]
    if wmat is None:
        wmat = np.identity(length)
    obj = cvxpy.Minimize(cvxpy.quad_form(moments-variables, wmat)) # objective function
    # obj = cvxpy.Minimize(cvxpy.sum_squares(x - moments)) 

    # the following gives constraints
    # Ref for PSD condition: [Lasserre 2009, Theorem 3.3 and 3.4]
    if length % 2 == 1:
        # odd case
        k = int((length+1)/2)
        h_mat = cvxpy.Variable((k, k+1))
        constraints = [h_mat[:, 1:]-interval[0]*h_mat[:, :k]>>0,
                       interval[1]*h_mat[:, :k]-h_mat[:, 1:]>>0]
    else:
        # even case
        k = int(length/2)+1
        h_mat = cvxpy.Variable((k, k))
        constraints = [h_mat>>0,
                       (interval[0]+interval[1])*h_mat[:k-1, 1:]-interval[0]*interval[1]*h_mat[:k-1, :k-1]-h_mat[1:, 1:]>>0]
    num_row, num_col = h_mat.shape
    for i in range(num_row):
        for j in range(num_col):
            if i == 0 and j == 0:
                constraints.append(h_mat[0, 0] == 1)
            else:
                constraints.append(h_mat[i, j] == variables[i+j-1])

    prob = cvxpy.Problem(obj, constraints)
    try:
        prob.solve(solver=cvxpy.CVXOPT)
    except Exception as e:
        warnings.warn("CVXOPT failed. Using SCS solver..."+str(e))
        prob.solve(solver=cvxpy.SCS)
        # prob.solve()
    # opt = prob.solve(solver=cvxpy.CVXOPT)

    return np.asarray(variables.value).reshape(moments.shape)




def quadmom(moments, dettol=0, inf=1e10):
    """ compute quadrature from moments
    ref: Gene Golub, John Welsch, Calculation of Gaussian Quadrature Rules

    Args:
    m: moments sequence
    dettol: tolerant of singularity of moments sequence (quantified by determinant of moment matrix)
    INF: infinity

    Returns:
    U: quadrature
    """

    moments = np.asarray(moments)
    # INF = float('inf')
    inf = 1e10

    if len(moments) % 2 == 1:
        moments = np.append(moments, inf)
    num = int(len(moments)/2)
    moments = np.insert(moments, 0, 1)


    h_mat = hankel(moments[:num+1:], moments[num::]) # Hankel matrix
    for i in range(len(h_mat)):
        # check positive definite and decide to use how many moments
        if np.linalg.det(h_mat[0:i+1, 0:i+1]) <= dettol: # alternative: less than some threshold
            h_mat = h_mat[0:i+1, 0:i+1]
            h_mat[i, i] = inf
            num = i
            break
    r_mat = np.transpose(np.linalg.cholesky(h_mat)) # upper triangular Cholesky factor

    # Compute alpha and beta from r, using Golub and Welsch's formula.
    alpha = np.zeros(num)
    alpha[0] = r_mat[0][1] / r_mat[0][0]
    for i in range(1, num):
        alpha[i] = r_mat[i][i+1]/r_mat[i][i] - r_mat[i-1][i]/r_mat[i-1][i-1]

    beta = np.zeros(num-1)
    for i in range(num-1):
        beta[i] = r_mat[i+1][i+1]/r_mat[i][i]

    jacobi = np.diag(alpha, 0) + np.diag(beta, 1) + np.diag(beta, -1)

    eigval, eigvec = np.linalg.eig(jacobi)

    atoms = eigval
    weights = moments[0] * np.power(eigvec[0], 2)

    return DiscreteRV(w=weights, x=atoms)


def deconvolve_unknown_variance(moments):
    """ Deconvolution with unknown sigma, using Lindsay's estimator.
    Fit moments with U+sigma Z. Estimate common sigma, and moments of U.

    Args:
    moments: array of float, length 2k
    moments estimate of degree 1 to 2k

    Returns:
    Tuple of deconvolved moments and estimated variance (sigma^2)
    """

    moments = np.insert(moments, 0, 1)

    length = len(moments)
    m_hermite = [0]*length
    x_var = np.poly1d([1, 0]) # x = sigma^2

    if length > 0:
        prepre = np.zeros(moments.shape)
        prepre[0]=1
        m_hermite[0] = moments[0]
    if length > 1:
        pre = np.zeros(moments.shape)
        pre[1]=1
        m_hermite[1] = moments[1]
    for k in range(2, length):
        # recursion: H_{n+1}(x) = x * H_n(x) - n * H_{n-1}(x)
        coeffs = np.roll(pre, 1) - prepre*(k-1)
        for i in range(k+1):
            m_hermite[k] += float(coeffs[i]*moments[i])*(x_var**(int((k-i)/2)))
        prepre = pre
        pre = coeffs

    # Solve the first non-negative root
    equation = det_mom(m_hermite)

    root = equation.r
    root = root[np.isreal(root)].real
    root = root[root >= 0]
    root.sort()
    # print(root)
    root0 = root[0]

    for k in range(2, length):
        m_hermite[k] = m_hermite[k](root0)

    return (np.asarray(m_hermite[1:]), float(root0))



def det_mom(moments):
    """
    Compute determinant of moment matrix

    Args:
    moments: array of moments of

    Return:
    determinant of moment matrix
    """
    return determinant(get_hankel(moments))


def get_hankel(moments):
    """
    Construct Hankel matrix from (m_0,...,m_{2k})

    Args:
    moments: array of floats of length 2k+1
    moments of degrees from 0 to 2k

    Return:
    Hankel matrix from those moments of size (k+1,k+1)
    """
    # length of m = 2k+1 = 2k_inc-1
    k_inc = int((len(moments)+1)/2)
    matrix = [[0]*k_inc for i in range(k_inc)]
    for i in range(k_inc):
        for j in range(k_inc):
            matrix[i][j] = moments[i+j]

    return matrix


def determinant(mat, rows=None, cols=None):
    """
    Compute the determinant of a submatrix

    Args:
    mat: matrix, list of lists
    input matrix

    rows: list of int
    selection of rows

    cols: list of int, same size as rows
    selection of columns

    Returns:
    determinant of submatrix mat[rows, cols]
    """
    if rows is None:
        num_rows, num_cols = len(mat), len(mat[0])
        rows = list(range(num_rows))
        cols = list(range(num_cols))

    num_rows = len(rows)
    num_cols = len(cols)
    assert num_rows == num_cols

    if num_rows == 1:
        return mat[rows[0]][cols[0]]
    # elif rows == 2:
    #     return M[r[0]][c[0]]*M[r[1]][c[1]] - M[r[0]][c[1]]*M[r[1]][c[0]]
    # elif rows == 3:
    #     return (M[r[0]][c[0]]*M[r[1]][c[1]]*M[r[2]][c[2]]
    #             + M[r[0]][c[1]]*M[r[1]][c[2]]*M[r[2]][c[0]]
    #             + M[r[0]][c[2]]*M[r[1]][c[0]]*M[r[2]][c[1]])
    #             - (M[r[0]][c[2]]*M[r[1]][c[1]]*M[r[2]][c[0]]
    #             + M[r[0]][c[0]]*M[r[1]][c[2]]*M[r[2]][c[1]]
    #             + M[r[0]][c[1]]*M[r[1]][c[0]]*M[r[2]][c[2]])
    else:
        det = 0
        newr = rows[1:]
        sign = 1
        for k in range(num_cols):
            newc = cols[:k] + cols[(k + 1):]
            det += determinant(mat, newr, newc)*mat[rows[0]][cols[k]]*sign
            sign *= -1
        return det

    #### Available methods for computing determinant using Sympy: bareis, berkowitz, det_LU
    #### Method 1: berkowitz
    # eq = sympy.Matrix(H).det(method='berkowitz').as_poly().expand()
    #### Method 2: bareis
    # eq = sympy.Matrix(H).det(method='bareis').as_poly()
    # for i in eq.gens[1:]:
    #     eq = eq.eval(i,1)
    # return eq



"""
    Wave.Scattering2.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import matplotlib.animation as animation
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar

class BeamAnimDialog(Dialog):
    def __init__(self, parent):
        Dialog.__init__(self, parent, "Beam Animatio", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):
        cframe = Tk.Frame(body_frame)
        cframe.pack()
        bframe = Tk.Frame(body_frame)
        bframe.pack(fill=Tk.BOTH, expand=1)
        tframe = Tk.Frame(bframe)
        tframe.pack(side=Tk.LEFT)
        pframe = Tk.Frame(bframe)
        pframe.pack(side=Tk.RIGHT)

        fig, axes = plt.subplots(ncols=2, figsize=(14, 7))
        self.fig  = fig
        self.mpl_canvas = FigureCanvasTkAgg(fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)
        self.toolbar = NavigationToolbar(self.mpl_canvas, tframe)
        self.toolbar.update()
        self.draw_anim(fig, axes)
        self.mpl_canvas.draw()

    def draw_anim(self, fig, axes):
        ax1, ax2 = axes

        for ax in axes:
            ax.grid(False)

        N = 800
        n = N//2
        r = np.arange(N)
        rr = r**2
        i = np.arange(N)
        ii, jj = np.meshgrid(i, i)
        labels1 = np.searchsorted(rr, (ii - n)**2 + (jj-n)**2)

        canvas = np.zeros((N, N))

        def color_filter(v, maxv):
            return 1/(1 + np.exp(-2*(v - (maxv-0.01))))

        cmap = 'plasma'
        target_particle1 = labels1 <= 5

        d = 15
        m = n - d
        labels2 = np.searchsorted(rr, (ii - m)**2 + (jj-m)**2)
        target_particle2 = labels2 <= 5

        Lmax = np.max([labels1, labels2])
        x = np.linspace(0, np.pi*96, N+Lmax)

        def animate_func(i):
            t = np.pi*i/2
            beam1 = color_filter(1 + np.sin(x[0:n] - t), 2)
            canvas[n-2:n+3,0:n] = beam1     # as if linewith=5
            beam2 = color_filter(1 + np.sin(x[0:m] - t), 2)
            canvas[m-2:m+3,0:m] = beam2     # as if linewith=5
            canvas[target_particle1] = 1
            canvas[target_particle2] = 1
            im1 = ax1.imshow(canvas, animated=True, cmap=cmap, origin='lower')
            scattered1 = 1 + np.sin(x[n:] - t)
            intensity1 = scattered1[labels1]
            scattered2 = 1 + np.sin(x[m:] - t)
            intensity2 = scattered2[labels2]
            im2 = ax2.imshow(color_filter(intensity1+intensity2, 4), animated=True, cmap=cmap, origin='lower')
            return [im1, im2]

        fig.tight_layout()
        self.ani = animation.FuncAnimation(fig, animate_func, frames=60, interval=50, blit=True)
        # self.ani.save('scattering.mp4', writer="ffmpeg")

def demo2():
    from molass_legacy.KekLib.TkUtils import get_tk_root

    root = get_tk_root()

    def show_demo():
        dialog = BeamAnimDialog(root)
        dialog.show()
        root.quit()

    root.after(100, show_demo)
    root.mainloop()
    root.destroy()



"""
    Solvers.UltrNest.SolverUltraNest.py

    Copyright (c) 2024, SAXS Team, KEK-PF    
"""
import logging
import numpy as np
from ultranest import ReactiveNestedSampler 
from ultranest.stepsampler import SliceSampler, generate_mixture_random_direction
from molass_legacy.Optimizer.OptimizerUtils import OptimizerResult
from molass_legacy.Optimizer.StateSequence import save_opt_params

NARROW_BIND_ALLOW = 1.0

def get_max_ncalls(niter):
    return niter*7000

class SolverUltraNest:
    def __init__(self, optimizer):
        self.optimizer = optimizer
        self.shm = optimizer.shm
        self.num_pure_components = optimizer.num_pure_components
        self.cb_fh = optimizer.cb_fh
        self.callback_counter = 0
        self.logger = logging.getLogger(__name__)

    def minimize(self, objective, init_params, niter=100, seed=1234, bounds=None, callback=None, narrow_bounds=True):
        from importlib import reload
        import Solvers.UltraNest.SamplerCallback
        reload(Solvers.UltraNest.SamplerCallback)
        from Solvers.UltraNest.SamplerCallback import SamplerCallback

        num_params = len(init_params)
        self.objective = objective

        if narrow_bounds:
            lower = init_params - NARROW_BIND_ALLOW
            upper = init_params + NARROW_BIND_ALLOW
        else:
            lower = bounds[:,0]
            upper = bounds[:,1]
        def my_prior_transform(cube):
            # transform location parameter: uniform prior
            params = cube * (upper - lower) + lower
            return params

        def my_likelihood(params):
            # print("objective_func_wrapper: par=", par)
            fv = objective(params)
            return -fv

        # logging.basicConfig(level=logging.INFO)     # to suppress debug log

        param_names = ["p%02d" % i for i in range(num_params)]
        sampler = ReactiveNestedSampler(param_names, my_likelihood, my_prior_transform)
        sampler.logger.setLevel(logging.INFO)       # to suppress debug log
        sampler_callback = SamplerCallback(self, sampler)

        self.logger.info("running without any step sampler")
        result1 = sampler.run(min_num_live_points=400, max_ncalls=10000, viz_callback=sampler_callback)

        self.logger.info("running with a step sampler: SliceSampler")
        # add a step sampler: from the "Higher-dimensional fitting" tutorial
        nsteps = 2 * num_params
        # create step sampler:
        sampler.stepsampler = SliceSampler(
            nsteps=nsteps,
            generate_direction=generate_mixture_random_direction,
            # adaptive_nsteps=False,
            # max_nsteps=400
        )

        max_ncalls = get_max_ncalls(niter)
        result2 = sampler.run(min_num_live_points=400, max_ncalls=max_ncalls, viz_callback=sampler_callback)

        opt_params = result2['maximum_likelihood']['point']

        return OptimizerResult(x=opt_params, nit=niter, nfev=self.optimizer.eval_counter)

    def callback(self, norm_params, f, accept):
        fv = self.objective(norm_params)
        self.logger.info("callback: fv=%.3g", fv)
        real_params = self.optimizer.to_real_params(norm_params)
        save_opt_params(self.cb_fh, real_params, fv, accept, self.optimizer.eval_counter)
        self.callback_counter += 1
        if self.shm is not None:
            self.shm.array[0] = self.callback_counter
        return False


"""

    OutlineFigure.py

    Copyright (c) 2018-2024, SAXS Team, KEK-PF

"""
import os
import sys
import numpy as np
import logging
import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg  import FigureCanvasTkAgg
import matplotlib.gridspec as gridspec
from molass_legacy.KekLib.OurTkinter import Tk
from molass_legacy.KekLib.TkUtils import split_geometry
from molass_legacy._MOLASS.SerialSettings import get_setting, get_xray_picking

ADD_TOOLBAR         = False
LEGEND_FONTSIZE     = 10
LEGEND_FONTSIZE_LARGE = 14

class OutlineFigure(Tk.Frame):
    def __init__(self, parent, dialog, *args, **kwargs):
        self.parent = parent
        self.dialog = dialog
        self.high_dpi = dialog.high_dpi
        self.text_font = 22 if self.high_dpi else 16
        self.use_average = kwargs.pop('use_average', False)
        self.file_info_table = kwargs.pop('file_info_table', None)
        figsize = kwargs.pop('figsize', (4,5))
        self.enable_usable_limit = get_setting('enable_usable_limit')
        self.absorbance_picking = get_setting('absorbance_picking')
        self.averaged_elution = None
        self.Rg = None
        self.xray_elution_restrict = None
        self.xray_angle_restrict = None
        self.uv_elution_restrict = None

        self.logger = logging.getLogger( __name__ )

        Tk.Frame.__init__(self, parent)

        self.fig = plt.figure( figsize=figsize )
        self.mpl_canvas = FigureCanvasTkAgg( self.fig, self )
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack( fill=Tk.BOTH, expand=1 )
        self.mpl_canvas.mpl_connect('button_press_event', self.on_mpl_button_press)

        if ADD_TOOLBAR:
            self.min_toolbar = Tk.Label(self)
            self.min_toolbar.pack()
            self.mpl_canvas.mpl_connect('motion_notify_event', self.on_mpl_motion_notify)

        gs = gridspec.GridSpec(2, 1)
        self.ax1 = self.fig.add_subplot(gs[0, 0])
        self.ax1.set_axis_off()
        self.ax2 = self.fig.add_subplot(gs[1, 0])
        self.ax2.set_axis_off()

        self.fig.tight_layout(pad=0, w_pad=0, h_pad=0, rect=[-0.05, 0, 1, 1])
        self.popup_menu = None

    def create_popup_menu( self ):
        # TODO: unify this menu and Tools menu
        from molass_legacy._MOLASS.Version import is_developing_version
        if self.popup_menu is None:
            self.popup_menu = Tk.Menu( self, tearoff=0 )
            self.popup_menu.add_command( label='Data Range Trimming', command=self.show_datarange_dialog )
            self.popup_menu.add_command( label="Remove Abnormal Points", command=self.show_abnomality_check_dialog, state=Tk.DISABLED )
            if is_developing_version():
                self.popup_menu.add_command(label='Save the upper figure', command=self.save_the_upper_figure)
                self.popup_menu.add_command(label='Save the XR elution curve', command=lambda: self.save_the_elution_curve(0))
                self.popup_menu.add_command(label='Save the UV elution curve', command=lambda: self.save_the_elution_curve(1))

    def clear_figures(self):
        for ax in [self.ax1, self.ax2]:
            ax.cla()
            ax.set_axis_off()

        self.mpl_canvas.draw()

    def draw_guide_message( self, message ):
        ax1 = self.ax1
        ax1.cla()
        ax1.set_axis_off()
        ax1.text(0.5, 0.5, message, ha='center', fontsize=self.text_font, alpha=0.2)
        self.mpl_canvas.draw()

    def set_data(self, pre_recog):
        self.pre_recog = pre_recog
        serial_data = pre_recog.get_pre_recog_copy()
        self.serial_data = serial_data
        self.similarity = pre_recog.cs
        self.averaged_elution = serial_data.ivector
        self.elution_size = len(serial_data.ivector)
        exact_index = ( serial_data.xray_slice.start + serial_data.xray_slice.stop ) //2
        self.exact_elution = serial_data.intensity_array[:,exact_index,1]
        self.q_ = serial_data.qvector
        self.default_eno = serial_data.xray_curve.primary_peak_i
        self.selected = self.default_eno
        self.xray_array = serial_data.intensity_array
        self.angle_size = self.xray_array.shape[1]
        self.usable_limit = serial_data.get_usable_q_limit()
        self.pre_rg = pre_recog.pre_rg
        use_xray_conc  = get_setting('use_xray_conc')
        use_mtd_conc  = get_setting('use_mtd_conc')
        if use_xray_conc == 0 and use_mtd_conc == 0:
            self.mapped_x = self.similarity.get_extended_x()
            self.mapped_y = self.similarity.get_uniformly_mapped_a_curve(x=self.mapped_x)
        else:
            self.mapped_x = None
            self.mapped_y = None

    def update_elution_curve(self):
        from bisect import bisect_right
        serial_data = self.serial_data
        # print(__name__, 'serial_data=', id(serial_data))
        q = get_xray_picking()
        exact_index = bisect_right(serial_data.qvector, q)
        self.exact_elution = serial_data.intensity_array[:,exact_index,1]
        self.draw_figure1()
        self.mpl_canvas.draw()

    def draw_figure(self, selected=None):
        # print('draw_figure', selected)
        if selected is None:
            selected = self.default_eno

        self.selected = selected
        self.get_restrict_info()
        self.draw_figure1()
        self.draw_figure2()
        self.draw_rg_text()
        self.mpl_canvas.draw()

    def get_restrict_info(self):
        uv_restrict_list = get_setting('uv_restrict_list')
        if uv_restrict_list is None:
            self.uv_elution_restrict = None
        else:
            rec = uv_restrict_list[0]
            self.uv_elution_restrict = rec if rec is None else rec.get_safe_object()

        xr_restrict_list = get_setting('xr_restrict_list')
        if xr_restrict_list is None:
            self.xray_elution_restrict = None
            self.xray_angle_restrict = None
        else:
            rec = xr_restrict_list[0]
            self.xray_elution_restrict = rec if rec is None else rec.get_safe_object()
            rec = xr_restrict_list[1]
            self.xray_angle_restrict = rec if rec is None else rec.get_safe_object()

    def draw_figure1(self):
        if self.averaged_elution is None:
            # before set_data?
            return

        self.draw_outline_elution(self.ax1)

    def draw_outline_elution(self, ax1):
        ax1.cla()
        # ax1.set_title("Xray elution curve")
        ax1.axes.get_yaxis().set_ticks([])

        if self.mapped_y is not None:
            ax1.plot( self.mapped_x, self.mapped_y, ':', color='blue', label='UV at wavelength=%.3g' % self.absorbance_picking )

        if self.use_average:
            label = 'average around Q=%.2g' % get_xray_picking()
            elution_y = self.averaged_elution
        else:
            label = 'Xray at Q=%.2g' % get_xray_picking()
            elution_y = self.exact_elution
        ax1.plot( elution_y, color='orange', label=label )
        ax1.plot( self.selected, elution_y[self.selected], 'o', color='yellow', markersize=5 )
        xmin, xmax = ax1.get_xlim()
        ymin, ymax = ax1.get_ylim()
        ax1.set_ylim(ymin, ymax)

        if self.uv_elution_restrict is not None:
            flag, lower, stop, _ = self.uv_elution_restrict
            uv_lower, uv_stop = [ self.similarity.inverse_int_value(j) for j in [lower, stop] ]

            if self.xray_elution_restrict is None:
                color = 'purple'
                alpha = 0.5
                lower, stop = uv_lower, uv_stop
            else:
                flag, xr_lower, xr_stop, _ = self.xray_elution_restrict
                color = 'black'
                alpha = 1.0
                # to avoid misunderstanding that the range is not appropriate
                # (may be afraid of being too wide as in 20230304)
                lower = max(uv_lower, xr_lower)
                stop = min(uv_stop, xr_stop)

            for x_ in [lower, stop-1 ]:
                ax1.plot( [x_, x_], [ymin, ymax], ':', color=color, alpha=alpha )

        fontsize = LEGEND_FONTSIZE_LARGE if self.high_dpi else LEGEND_FONTSIZE
        ax1.legend(fontsize=fontsize)

    def draw_figure2(self):
        ax2 = self.ax2
        ax2.cla()
        ax2.axes.get_yaxis().set_ticks([])
        ax2.xaxis.set_major_formatter(plt.FuncFormatter(self.format_func))
        if self.use_average:
            label = 'Log average near elution No. %d' % self.selected
            scattering_y = self.pre_rg.get_scattering_y(self.selected, slice(None, None))
        else:
            label = 'Log $I$ at elution No. %d' % self.selected
            scattering_y = self.xray_array[self.selected, :, 1]

        ax2.plot(self.q_, np.log10(scattering_y), label=label)

        if self.xray_angle_restrict is not None:
            ymin, ymax = ax2.get_ylim()
            ax2.set_ylim(ymin, ymax)
            flag, start, stop, _ = self.xray_angle_restrict
            for i in [start, stop - 1]:
                if i is None:
                    continue
                x_ = self.q_[i]
                ax2.plot( [x_, x_], [ymin, ymax], ':', color='black' )

        fontsize = LEGEND_FONTSIZE_LARGE if self.high_dpi else LEGEND_FONTSIZE
        ax2.legend(loc='upper right', fontsize=fontsize)

    def format_func(self, value, tick_number):
        return '%.2g' % (value)

    def on_mpl_button_press(self, event):
        if event.xdata is None or self.averaged_elution is None:
            return

        if event.button == 3:
            self.create_popup_menu()
            w, h, x, y = split_geometry(self.dialog.geometry())
            self.popup_menu.post(x + event.x + 30, y + h - event.y)
            return

        if event.inaxes != self.ax1:
            return

        min_d = None
        min_x = None
        for x, y in enumerate(self.averaged_elution):
            d = (x - event.xdata)**2 + (y - event.ydata)**2
            if min_d is None or d < min_d:
                min_d = d
                min_x = x
        if min_x is not None:
            # print('min_x=', min_x)
            if self.file_info_table is not None:
                # as in the case of unit test
                self.file_info_table.select_row(min_x)
            self.draw_figure(selected=min_x)

    def on_mpl_motion_notify(self, event):
        if event.xdata is None:
            text = ''
        else:
            text = 'x=%d' % event.xdata
        self.min_toolbar.config(text=text)

    def compute_rg(self):
        return self.pre_rg.compute_rg(self.selected)

    def draw_rg_text(self, recompute=True):
        ax2 = self.ax2
        xmin, xmax = ax2.get_xlim()
        ymin, ymax = ax2.get_ylim()

        if recompute:
            Rg = self.Rg = self.compute_rg()
        else:
            Rg = self.Rg
        Rg_ = 'None' if Rg is None else '%.1f' % Rg
        qmax_i = len(self.q_)-1 if self.xray_angle_restrict is None else self.xray_angle_restrict.stop - 1
        qmRg = 'None' if Rg is None else '%.1f' % (Rg * self.q_[qmax_i])
        tx = xmin * 0.95 + xmax * 0.05
        ty = ymin * 0.95 + ymax * 0.05
        ax2.text(tx, ty, '$R_g$=%s\n$Q_{max} \\times R_g$=%s' % (Rg_, qmRg), alpha=0.2, fontsize=self.text_font )

        cx, cy = self.q_[qmax_i], ymin
        xoffset = (xmax - xmin) * 0.05
        yoffset = (ymax - ymin) * 0.15

        if qmax_i/len(self.q_) < 0.75:
            xoffset_ = xoffset
            ha = 'left'
        else:
            xoffset_ = -xoffset
            ha = 'right'

        ax2.annotate( '$Q_{max}$', xy=(cx, cy ),
                    xytext=( cx + xoffset_, cy + yoffset ), alpha=0.5,
                    arrowprops=dict( headwidth=3, headlength=8, width=0.5, color='black', alpha=0.5),
                    ha=ha, va='center', fontsize=self.text_font
                    )

    def show_datarange_dialog(self):
        print('show_datarange_dialog')
        from molass_legacy.Trimming import DataRangeDialog
        dialog = DataRangeDialog(self.dialog, self.pre_recog)
        dialog.show()
        self.dialog.draw_figure()

    def show_abnomality_check_dialog(self):
        pass

    def save_the_figure( self, folder, analysis_name ):
        from DataUtils import cut_upper_folders
        in_folder = cut_upper_folders(get_setting('in_folder'))
        self.fig.suptitle(in_folder)
        self.fig.subplots_adjust(top=0.9)

        # print( 'save_the_figure: ', folder, analysis_name )
        filename = analysis_name.replace( 'analysis', 'figure' )
        path = os.path.join( folder, filename )
        self.fig.savefig( path )

    def save_the_upper_figure(self):
        from importlib import reload
        import Trimming.OutlineFigureSaver
        reload(Trimming.OutlineFigureSaver)
        from molass_legacy.Trimming.OutlineFigureSaver import save_the_upper_figure_impl
        save_the_upper_figure_impl(self)

    def save_the_elution_curve(self, type):
        from importlib import reload
        import Trimming.OutlineFigureSaver
        reload(Trimming.OutlineFigureSaver)
        from molass_legacy.Trimming.OutlineFigureSaver import save_the_elution_curve_impl
        save_the_elution_curve_impl(self, type)



# coding: utf-8
"""
    MctRangeInfo.py

    Copyright (c) 2019-2020, SAXS Team, KEK-PF
"""

class MctRangeInfo:
    def __init__(self, conc, ranges, curves):
        self.conc = conc
        self.ranges = ranges
        self.curves = curves



"""
    Trimming.SigmoidGuessX0.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
import ruptures as rpt
import molass_legacy.KekLib.DebugPlot as plt

def guess_x0_impl(x, y, debug=False):

    algo = rpt.Dynp(model="l2").fit(y)
    breakpoints = algo.predict(n_bkps=4)
    i = breakpoints[0]

    if debug:
        print("breakpoints=", breakpoints)
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.set_title("guess_x0_impl")
            ax.plot(x, y, label="data")
            ymin, ymax = ax.get_ylim()
            ax.set_ylim(ymin, ymax)
            x0 = x[i]
            ax.plot([x0, x0], [ymin, ymax], ":", lw=2, color="red", label="x0")
            for j in breakpoints[:-1]:
                ax.axvline(x=x[j], color='yellow')
            fig.tight_layout()
            plt.show()

    return i


"""

    SummaryExcelFormatter.py

    Copyright (c) 2023, SAXS Team, KEK-PF

"""
from ExcelCOM import xlCategory, xlValue, xlLow, msoThemeColorText1
from ExcelLogChart import ExcelLogChart
from .DefaultFont import set_default_bookfont_com
from .CharacterFormatter import to_superscript_font, to_subscript_font, to_italic_font, to_italic_subscipt_font

class SummaryArgs:
    def __init__(self, title, book_file):
        self.title = title
        self.book_file = book_file

def add_summary_format_setting(excel_client, args, logger):
    title = args.title
    book_file = args.book_file
    book_path = "%s(%s)" % (book_file, title)
    logger.info("add_summary_format_setting start for %s", book_path)

    ws = excel_client.openWorksheet(book_path)
    set_default_bookfont_com(ws.workbook)

    for row in [7, 8, 11, 13, 14, 23, 42, 44, 48, 51, 53]:
        cell = ws.worksheet.Cells(row,1)
        to_superscript_font(cell)

    for row in [24]:    # H₂O
        cell = ws.worksheet.Cells(row,2)
        to_subscript_font(cell)

    for row in [23, 35, 42, 44, 45, 46, 47, 48, 51, 52, 54]:
        cell = ws.worksheet.Cells(row,1)
        to_italic_font(cell, ["Q", "I(0)", "P(r)", "Q×", "M"])

    for row in [43, 45, 49, 50]:
        cell = ws.worksheet.Cells(row,1)
        to_italic_subscipt_font(cell, ["Rg", "Dmax"])

    ws.workbook.Save()
    ws.workbook.Close()     # required in the more_multicore mode



# coding: utf-8
"""

    ファイル名：   DebugCanvas.py

    処理内容：

        デバッグ用の plot 

    Copyright (c) 2017, Masatsuyo Takahashi, KEK-PF

"""
import numpy                            as np
import matplotlib.pyplot                as plt
from matplotlib.backends.backend_tkagg  import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter                         import Tk, Dialog
from molass_legacy.KekLib.TkUtils                            import adjusted_geometry

class DebugCanvas( Dialog ):
    def __init__( self, title, draw_func, parent=None, figsize=None, toolbar=False ):

        if parent is None:
            parent = Tk.Toplevel()
            parent.withdraw()
            self.created_parent = True
        else:
            self.created_parent = False

        self.grab = 'local'     # used in grab_set
        self.parent     = parent
        self.title_     = title
        self.figsize    = figsize
        self.draw_func  = draw_func
        self.continue_  = None
        self.toolbar    = toolbar

    def show( self, cancelable=False, cunstom_button_cb=None, cursor_update=True ):
        self.cancelable = cancelable
        self.cunstom_button_cb = cunstom_button_cb
        self.cursor_update = cursor_update
        # self.parent.after( 0, self.adjust_geometry )
        Dialog.__init__( self, self.parent, self.title_ )
        # TODO: self.resizable(width=False, height=False)

    def destroy_parent( self ):
        if self.created_parent:
            self.parent.destroy()

    def body( self, body_frame ):   # overrides parent class method

        cframe = Tk.Frame( body_frame )
        cframe.pack()

        figsize_ = ( 18, 9 ) if self.figsize is None else self.figsize

        fig = plt.figure( figsize=figsize_ )

        self.mpl_canvas = FigureCanvasTkAgg( fig, cframe )
        self.mpl_canvas.draw()
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack( fill=Tk.BOTH, expand=1 )

        if self.toolbar:
            from molass_legacy.KekLib.OurMatplotlib  import NavigationToolbar
            self.toolbar_widget = NavigationToolbar( self.mpl_canvas, cframe )

        # it seems that draw_func should be called after the creation of mpl_canvas
        # in order to enable 3d-rotation of the figure
        self.draw_func( fig )
        if self.cursor_update:
            self.parent.config( cursor='' )

        self.protocol( "WM_DELETE_WINDOW", self.ok )

    def buttonbox(self):
        '''add standard button box.

        override if you do not want the standard buttons
        '''

        if self.cunstom_button_cb is not None:
            self.cunstom_button_cb( self )
            return

        box = Tk.Frame(self)

        w = Tk.Button(box, text="OK", width=10, command=self.ok, default=Tk.ACTIVE)
        w.pack(side=Tk.LEFT, padx=5, pady=5)

        if self.cancelable:
            w = Tk.Button(box, text="Cancel", width=10, command=self.cancel)
            w.pack(side=Tk.LEFT, padx=5, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

        box.pack()

    def ok( self ):
        self.destroy_parent()
        self.continue_  = True
        self.destroy()

    def cancel( self ):
        self.destroy_parent()
        self.continue_  = False
        self.destroy()

    def adjust_geometry( self ):
        self.update()
        # TODO: fix being placed to the exceeding left
        self.geometry( adjusted_geometry( self.geometry() ) )




# coding: utf-8
"""
    BaseSurface.py

    Copyright (c) 2019, SAXS Team, KEK-PF
"""
import logging
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt
from MatrixData import simple_plot_3d
from molass_legacy.Baseline.ScatteringBaseline import ScatteringBaseline
from ScatteringBasecurve import ScatteringBasecurve
from ScatteringBasespline import ScatteringBasespline
from molass_legacy.KekLib.SciPyCookbook import smooth
from ThreeDimUtils import compute_plane
from molass_legacy.Baseline.LambertBeer import BasePlane

class BaseSurface:
    def __init__(self, xd):
        self.logger = logging.getLogger(__name__)
        data, vector = xd.get_sliced_data()
        self.data = data
        self.vector = vector
        self.e_curve = xd.e_curve  # TODO
        self.xd = xd

        ends_array, ends_array_smoothed = self.compute_lpm_baselines()
        base_plane = self.lines_to_surface(ends_array)
        base_plane_smoothed = self.lines_to_surface(ends_array_smoothed)

        corrected_data = data - base_plane_smoothed

        index = xd.e_index    # TODO
        j0 = xd.j_slice.start
        bp = BasePlane(corrected_data, index, self.e_curve, j0=0 if j0 is None else j0)
        a, b, c = bp.solve()
        x = bp.x
        y = bp.y
        base_plane_adjusted  = compute_plane(a, b, c, x, y)
        adjusted_data = corrected_data - base_plane_adjusted

        if True:
            from DataUtils import get_in_folder
            plt.push()
            fig = plt.figure()
            ax = fig.add_subplot(111, projection='3d')
            ax.set_title("Non-Corrected Data", fontsize=20)
            simple_plot_3d(ax, xd.data, x=xd.vector, alpha=0.3)
            fig.tight_layout()
            plt.show()
            plt.pop()

            fig = plt.figure(figsize=(21,11))
            fig.suptitle("Correction using LPM+MF Base Surface for " + get_in_folder(), fontsize=20)
            ax1 = fig.add_subplot(231, projection='3d')
            ax2 = fig.add_subplot(232, projection='3d')
            ax3 = fig.add_subplot(233, projection='3d')
            ax4 = fig.add_subplot(234, projection='3d')
            ax5 = fig.add_subplot(235, projection='3d')
            ax6 = fig.add_subplot(236, projection='3d')

            # xd.plot(ax, color='yellow', alpha=0.1)
            ax1.set_title("Corrected with Base Lines", fontsize=20)
            ax2.set_title("Corrected with Base Surface", fontsize=20)
            ax3.set_title("Adjusted with Matrix Factorization", fontsize=20)
            simple_plot_3d(ax1, data - base_plane, x=vector, alpha=0.3)
            simple_plot_3d(ax2, corrected_data, x=vector, alpha=0.3)
            simple_plot_3d(ax3, adjusted_data, x=vector, alpha=0.3)
            ax4.set_title("LPM Base Lines", fontsize=20)
            ax5.set_title("Smoothed Base Surface", fontsize=20)
            ax6.set_title("Matrix Factorization-Adjust Plane", fontsize=20)
            simple_plot_3d(ax4, ends_array, x=vector, color='red', alpha=0.3)
            simple_plot_3d(ax5, base_plane_smoothed, x=vector, color='red', alpha=0.3)
            adjusted_plane = compute_plane(a, b, c, x[[0,-1]], y[[0,-1]])
            simple_plot_3d(ax6, base_plane_smoothed, x=vector, color='red', alpha=0.1)
            simple_plot_3d(ax6, adjusted_plane, x=vector[[0,-1]], y=y[[0,-1]], color='blue', alpha=0.3)
            fig.tight_layout()
            fig.subplots_adjust(top=0.92)
            plt.show()

    def compute_lpm_baselines(self):
        end_points_list = []
        y = np.array([0, self.data.shape[1]-1])
        for i in range(self.data.shape[0]):
            z = self.data[i,:]
            sbl = ScatteringBaseline( z, curve=self.e_curve, logger=self.logger )
            D, E  = sbl.solve()     # p_final=p_final
            end_points_list.append(D*y + E)
        work_array = np.array(end_points_list)
        work_array_smoothed = np.vstack([smooth(work_array[:,j], window_len=40) for j in range(2)]).T
        return work_array, work_array_smoothed

    def lines_to_surface(self, lines):
        w = np.linspace(0, 1, self.data.shape[1])
        W = np.vstack([1-w, w])
        return np.dot(lines, W)
